import os
from collections.abc import Iterator
from typing import Optional, Union

import numpy as np
import pandas as pd
from sklearn.utils._param_validation import StrOptions, validate_params

from skfp.datasets.utils import fetch_splits

from .bace import load_bace
from .bbbp import load_bbbp
from .clintox import load_clintox
from .esol import load_esol
from .freesolv import load_freesolv
from .hiv import load_hiv
from .lipophilicity import load_lipophilicity
from .muv import load_muv
from .pcba import load_pcba
from .sider import load_sider
from .tox21 import load_tox21
from .toxcast import load_toxcast


@validate_params(
    {
        "subset": [
            None,
            StrOptions(
                {
                    "classification",
                    "classification_single_task",
                    "classification_multitask",
                    "classification_no_pcba",
                    "regression",
                }
            ),
            list,
        ],
        "data_dir": [None, str, os.PathLike],
        "as_frame": ["boolean"],
        "verbose": ["boolean"],
    },
    prefer_skip_nested_validation=True,
)
def load_moleculenet_benchmark(
    subset: Optional[Union[str, list[str]]] = None,
    data_dir: Optional[Union[str, os.PathLike]] = None,
    as_frames: bool = False,
    verbose: bool = False,
) -> Union[
    Iterator[tuple[str, pd.DataFrame]], Iterator[tuple[str, list[str], np.ndarray]]
]:
    """
    Load and return the MoleculeNet benchmark datasets.

    Datasets have varied molecular property prediction tasks: regression, single-task,
    and multitask classification. Scaffold split is recommended for all of them,
    following Open Graph Benchmark [1]_. They differ in recommended metrics. For more
    details, see loading functions for particular datasets.

    Often only a subset of those datasets is used for benchmarking, e.g. only
    single-task datasets, or only classification datasets and excluding PCBA (due to its
    large size). A subset of datasets can be selected by using ``subset`` argument.

    Dataset names are also returned (case-sensitive). Datasets, grouped by task, are:

    - regression: ESOL, FreeSolv, Lipophilicity
    - single-task classification: BACE, BBBP, HIV
    - multitask classification: ClinTox, MUV, SIDER, Tox21, ToxCast, PCBA

    Parameters
    ----------
    subset : {None, "regression", "classification", "classification_single_task",
              "classification_multitask", "classification_no_pcba"} or list of strings
        If ``None``, returns all datasets. String loads only a given subset of all
        datasets. List of strings loads only datasets with given names.

    data_dir : {None, str, path-like}, default=None
        Path to the root data directory. If ``None``, currently set scikit-learn directory
        is used, by default `$HOME/scikit_learn_data`.

    as_frames : bool, default=False
        If True, returns the raw DataFrame for each dataset. Otherwise, returns SMILES
        as a list of strings, and labels as a NumPy array for each dataset.

    verbose : bool, default=False
        If True, progress bar will be shown for downloading or loading files.

    Returns
    -------
    data : generator of pd.DataFrame or tuples (list[str], np.ndarray)
        Loads and returns datasets with a generator. Returned types depend on the
        ``as_frame`` parameter, either:
        - Pandas DataFrame with columns: "SMILES", "label"
        - tuple of: list of strings (SMILES), NumPy array (labels)

    References
    ----------
    .. [1] `Zhenqin Wu et al.
        "MoleculeNet: a benchmark for molecular machine learning"
        Chem. Sci., 2018,9, 513-530
        <https://pubs.rsc.org/en/content/articlelanding/2018/sc/c7sc02664a>`_
    """
    regression_names = ["ESOL", "FreeSolv", "Lipophilicity"]
    clf_single_task_names = ["BACE", "BBBP", "HIV"]
    clf_multitask_names = ["ClinTox", "MUV", "SIDER", "Tox21", "ToxCast"]
    clf_pcba = ["PCBA"]

    dataset_name_to_func = {
        "ESOL": load_esol,
        "FreeSolv": load_freesolv,
        "Lipophilicity": load_lipophilicity,
        "BACE": load_bace,
        "BBBP": load_bbbp,
        "HIV": load_hiv,
        "ClinTox": load_clintox,
        "MUV": load_muv,
        "SIDER": load_sider,
        "Tox21": load_tox21,
        "ToxCast": load_toxcast,
        "PCBA": load_pcba,
    }

    if subset is None:
        dataset_names = (
            regression_names + clf_single_task_names + clf_multitask_names + clf_pcba
        )
    elif subset == "regression":
        dataset_names = regression_names
    elif subset == "classification":
        dataset_names = clf_single_task_names + clf_multitask_names + clf_pcba
    elif subset == "classification_single_task":
        dataset_names = clf_single_task_names
    elif subset == "classification_multitask":
        dataset_names = clf_multitask_names
    elif subset == "classification_no_pcba":
        dataset_names = clf_single_task_names + clf_multitask_names
    elif isinstance(subset, (list, set, tuple)):
        for name in subset:
            if name not in dataset_name_to_func:
                raise ValueError(
                    f"Dataset name '{name}' not recognized among MoleculeNet datasets"
                )
        dataset_names = subset
    else:
        raise ValueError(
            f'Value "{subset}" for subset not recognized, must be one of: '
            f'"classification", "classification_single_task", '
            f'"classification_no_pcba", "regression"; alternatively, it can'
            f"be a list of strings with dataset names from MoleculeNet to load"
        )

    dataset_functions = [dataset_name_to_func[name] for name in dataset_names]

    if as_frames:
        # generator of tuples (dataset_name, DataFrame)
        datasets = (
            (dataset_name, load_function(data_dir, as_frame=True, verbose=verbose))
            for dataset_name, load_function in zip(dataset_names, dataset_functions)
        )
    else:
        # generator of tuples (dataset_name, SMILES, y)
        datasets = (
            (dataset_name, *load_function(data_dir, as_frame=False, verbose=verbose))
            for dataset_name, load_function in zip(dataset_names, dataset_functions)
        )

    return datasets


@validate_params(
    {
        "dataset_name": [
            StrOptions(
                {
                    "ESOL",
                    "FreeSolv",
                    "Lipophilicity",
                    "BACE",
                    "BBBP",
                    "HIV",
                    "ClinTox",
                    "MUV",
                    "SIDER",
                    "Tox21",
                    "ToxCast",
                    "PCBA",
                }
            )
        ],
        "data_dir": [None, str, os.PathLike],
        "as_frame": ["boolean"],
        "verbose": ["boolean"],
    },
    prefer_skip_nested_validation=True,
)
def load_ogb_splits(
    dataset_name: str,
    data_dir: Optional[Union[str, os.PathLike]] = None,
    as_dict: bool = False,
    verbose: bool = False,
) -> Union[tuple[list[int], list[int], list[int]], dict[str, list[int]]]:
    """
    Load and return the MoleculeNet dataset splits from Open Graph Benchmark (OGB).

    OGB [1]_ uses precomputed scaffold split with 80/10/10% split between train/valid/test
    subsets. Test set consists of the smallest scaffold groups, and follows MoleculeNet
    paper [2]_. Those splits are widely used in literature.

    Dataset names here are the same as returned by `load_moleculenet_benchmark` function,
    and are case-sensitive.

    Parameters
    ----------
    dataset_name : {"ESOL", "FreeSolv", "Lipophilicity","BACE", "BBBP", "HIV", "ClinTox",
        "MUV", "SIDER", "Tox21", "ToxCast", "PCBA"}
        Name of the dataset to loads splits for.

    data_dir : {None, str, path-like}, default=None
        Path to the root data directory. If ``None``, currently set scikit-learn directory
        is used, by default `$HOME/scikit_learn_data`.

    as_dict : bool, default=False
        If True, returns the splits as dictionary with keys "train", "valid" and "test",
        and index lists as values. Otherwise, returns three lists with splits indexes.

    verbose : bool, default=False
        If True, progress bar will be shown for downloading or loading files.

    Returns
    -------
    data : tuple(list[int], list[int], list[int]) or dict
        Depending on the `as_dict` argument, one of:
        - three lists of integer indexes
        - dictionary with "train", "valid" and "test" keys, and values as lists with
        splits indexes

    References
    ----------
    .. [1] `Hu, Weihua, et al.
        "Open Graph Benchmark: Datasets for Machine Learning on Graphs."
        Advances in Neural Information Processing Systems 33 (2020): 22118-22133.
        <https://papers.neurips.cc/paper/2020/file/fb60d411a5c5b72b2e7d3527cfc84fd0-Paper.pdf>`_

    .. [2] `Zhenqin Wu et al.
        "MoleculeNet: a benchmark for molecular machine learning"
        Chem. Sci., 2018,9, 513-530
        <https://pubs.rsc.org/en/content/articlelanding/2018/sc/c7sc02664a>`_
    """
    splits = fetch_splits(
        data_dir,
        dataset_name=f"MoleculeNet_{dataset_name}",
        filename=f"ogb_splits_{dataset_name.lower()}.json",
        verbose=verbose,
    )
    if as_dict:
        return splits
    else:
        return splits["train"], splits["valid"], splits["test"]



from collections.abc import Sequence
from numbers import Integral
from typing import Any, Optional, Union

from rdkit.Chem import Mol
from rdkit.Chem.rdFingerprintGenerator import GetMorganGenerator
from rdkit.SimDivFilters.rdSimDivPickers import MaxMinPicker
from sklearn.utils._param_validation import Interval, RealNotInt, validate_params

from skfp.model_selection.splitters.utils import (
    ensure_nonempty_subset,
    split_additional_data,
    validate_train_test_split_sizes,
    validate_train_valid_test_split_sizes,
)
from skfp.utils.functions import get_data_from_indices
from skfp.utils.validators import ensure_mols


@validate_params(
    {
        "data": ["array-like"],
        "additional_data": ["tuple"],
        "train_size": [
            Interval(RealNotInt, 0, 1, closed="neither"),
            Interval(Integral, 1, None, closed="left"),
            None,
        ],
        "test_size": [
            Interval(RealNotInt, 0, 1, closed="neither"),
            Interval(Integral, 1, None, closed="left"),
            None,
        ],
        "return_indices": ["boolean"],
    },
    prefer_skip_nested_validation=True,
)
def maxmin_train_test_split(
    data: Sequence[Union[str, Mol]],
    *additional_data: Sequence,
    train_size: Optional[float] = None,
    test_size: Optional[float] = None,
    return_indices: bool = False,
    random_state: int = 0,
) -> Union[
    tuple[
        Sequence[Union[str, Mol]], Sequence[Union[str, Mol]], Sequence[Sequence[Any]]
    ],
    tuple[Sequence, ...],
    tuple[Sequence[int], Sequence[int]],
]:
    """
    Split using MaxMin algorithm.

    MaxMinPicker is an efficient algorithm for picking an optimal subset of diverse
    compounds from a candidate pool. The original algorithm was introduced in [1]_,
    but here we use an optimized implementation by Roger Sayle [2]_ [3]_ [4]_.

    First, molecules are vectorized using binary ECFP4 fingerprint (radius 2) with
    2048 bits. The first test molecule is picked randomly. Each next one is selected
    to maximize the minimal distance to the already selected molecules (hence the
    MaxMin name) [4]_. Distances are calculated on the fly as required.

    First, the test set is constructed, and training set are all other molecules.

    If ``train_size`` and ``test_size`` are integers, they must sum up to the ``data``
    length. If they are floating numbers, they must sum up to 1.

    Parameters
    ----------
    data : sequence
        A sequence representing either SMILES strings or RDKit ``Mol`` objects.

    additional_data: list[sequence]
        Additional sequences to be split alongside the main data (e.g., labels or feature vectors).

    train_size : float, default=None
        The fraction of data to be used for the train subset. If None, it is set to 1 - test_size.
        If test_size is also None, it will be set to 0.8.

    test_size : float, default=None
        The fraction of data to be used for the test subset. If None, it is set to 1 - train_size.
        If train_size is also None, it will be set to 0.2.

    return_indices : bool, default=False
        Whether the method should return the input object subsets, i.e. SMILES strings
        or RDKit ``Mol`` objects, or only the indices of the subsets instead of the data.

    Returns
    ----------
    subsets : tuple[list, list, ...]
        Tuple with train-test subsets of provided arrays. First two are lists of SMILES
        strings or RDKit ``Mol`` objects, depending on the input type. If `return_indices`
        is True, lists of indices are returned instead of actual data.

    References
    ----------
    .. [1] `Mark Ashton et al.
        "Identification of Diverse Database Subsets using Property-Based and Fragment-Based Molecular Descriptions"
        Quant. Struct.-Act. Relat., 21: 598-604
        <https://onlinelibrary.wiley.com/doi/10.1002/qsar.200290002>_`

    .. [2] `Roger Sayle
        "Improved RDKit implementation"
        <https://github.com/rdkit/UGM_2017/blob/master/Presentations/Sayle_RDKitDiversity_Berlin17.pdf>_`

    .. [3] `Tim Dudgeon
        "Revisting the MaxMinPicker"
        <https://rdkit.org/docs/cppapi/classRDPickers_1_1MaxMinPicker.html>_`

    .. [4] `Squonk - RDKit MaxMin Picker
        <https://squonk.it/docs/cells/RDKit%20MaxMin%20Picker>_`
    """
    data_size = len(data)
    train_size, test_size = validate_train_test_split_sizes(
        train_size, test_size, data_size
    )

    mols = ensure_mols(data)
    fps = GetMorganGenerator(radius=2, fpSize=2048).GetFingerprints(mols)

    picker = MaxMinPicker()
    test_idxs = picker.LazyBitVectorPick(
        fps,
        poolSize=data_size,
        pickSize=test_size,
        seed=random_state,
    )
    train_idxs = list(set(range(data_size)) - set(test_idxs))

    ensure_nonempty_subset(train_idxs, "train")
    ensure_nonempty_subset(test_idxs, "test")

    if return_indices:
        train_subset = train_idxs
        test_subset = test_idxs
    else:
        train_subset = get_data_from_indices(data, train_idxs)
        test_subset = get_data_from_indices(data, test_idxs)

    if additional_data:
        additional_data_split: list[Sequence[Any]] = split_additional_data(
            list(additional_data), train_idxs, test_idxs
        )
        return train_subset, test_subset, additional_data_split
    else:
        return train_subset, test_subset


@validate_params(
    {
        "data": ["array-like"],
        "additional_data": ["tuple"],
        "train_size": [
            Interval(RealNotInt, 0, 1, closed="neither"),
            Interval(Integral, 1, None, closed="left"),
            None,
        ],
        "valid_size": [
            Interval(RealNotInt, 0, 1, closed="neither"),
            Interval(Integral, 1, None, closed="left"),
            None,
        ],
        "test_size": [
            Interval(RealNotInt, 0, 1, closed="neither"),
            Interval(Integral, 1, None, closed="left"),
            None,
        ],
        "return_indices": ["boolean"],
    },
    prefer_skip_nested_validation=True,
)
def maxmin_train_valid_test_split(
    data: Sequence[Union[str, Mol]],
    *additional_data: Sequence,
    train_size: Optional[float] = None,
    valid_size: Optional[float] = None,
    test_size: Optional[float] = None,
    return_indices: bool = False,
    random_state: int = 0,
) -> Union[
    tuple[
        Sequence[Union[str, Mol]], Sequence[Union[str, Mol]], Sequence[Sequence[Any]]
    ],
    tuple[Sequence, ...],
    tuple[Sequence[int], Sequence[int]],
]:
    """
    Split using MaxMin algorithm.

    MaxMinPicker is an efficient algorithm for picking an optimal subset of diverse
    compounds from a candidate pool. The original algorithm was introduced in [1]_,
    but here we use an optimized implementation by Roger Sayle [2]_ [3]_ [4]_.

    First, molecules are vectorized using binary ECFP4 fingerprint (radius 2) with
    2048 bits. The first test molecule is picked randomly. Each next one is selected
    to maximize the minimal distance to the already selected molecules (hence the
    MaxMin name) [4]_. Distances are calculated on the fly as required.

    First, the test set is constructed, then validation, and training set are all
    other molecules.

    If ``train_size``, ``valid_size`` and ``test_size`` are integers, they must sum up
    to the ``data`` length. If they are floating numbers, they must sum up to 1.

    Parameters
    ----------
    data : sequence
        A sequence representing either SMILES strings or RDKit ``Mol`` objects.

    additional_data: sequence
        Additional sequences to be split alongside the main data, e.g. labels.

    train_size : float, default=None
        The fraction of data to be used for the train subset. If None, it is set
        to 1 - test_size - valid_size. If valid_size is not provided, train_size
        is set to 1 - test_size. If train_size, test_size and valid_size aren't
        set, train_size is set to 0.8.

    valid_size : float, default=None
        The fraction of data to be used for the test subset. If None, it is set
        to 1 - train_size - valid_size. If train_size, test_size and valid_size
        aren't set, train_size is set to 0.1.

    test_size : float, default=None
        The fraction of data to be used for the validation subset. If None, it is
        set to 1 - train_size - valid_size. If valid_size is not provided, test_size
        is set to 1 - train_size. If train_size, test_size and valid_size aren't set,
        test_size is set to 0.1.

    return_indices : bool, default=False
        Whether the method should return the input object subsets, i.e. SMILES strings
        or RDKit ``Mol`` objects, or only the indices of the subsets instead of the data.

    Returns
    ----------
    subsets : tuple[list, list, ...]
        Tuple with train-valid-test subsets of provided arrays. First three are lists of
        SMILES strings or RDKit ``Mol`` objects, depending on the input type. If
        `return_indices` is True, lists of indices are returned instead of actual data.

    References
    ----------
    .. [1] `Mark Ashton et al.
        "Identification of Diverse Database Subsets using Property-Based and Fragment-Based Molecular Descriptions"
        Quant. Struct.-Act. Relat., 21: 598-604
        <https://onlinelibrary.wiley.com/doi/10.1002/qsar.200290002>_`

    .. [2] `Roger Sayle
        "Improved RDKit implementation"
        <https://github.com/rdkit/UGM_2017/blob/master/Presentations/Sayle_RDKitDiversity_Berlin17.pdf>_`

    .. [3] `Tim Dudgeon
        "Revisting the MaxMinPicker"
        <https://rdkit.org/docs/cppapi/classRDPickers_1_1MaxMinPicker.html>_`

    .. [4] `Squonk - RDKit MaxMin Picker
        <https://squonk.it/docs/cells/RDKit%20MaxMin%20Picker>_`
    """

    data_size = len(data)
    train_size, valid_size, test_size = validate_train_valid_test_split_sizes(
        train_size, valid_size, test_size, len(data)
    )
    mols = ensure_mols(data)
    fps = GetMorganGenerator(radius=2, fpSize=2048).GetFingerprints(mols)

    picker = MaxMinPicker()

    # select the test set only
    test_idxs = picker.LazyBitVectorPick(
        fps,
        poolSize=data_size,
        pickSize=test_size,
        seed=random_state,
    )

    # select validation + test sets, first including already computed test set
    # then remove test indexes, leaving only validation set
    valid_idxs = picker.LazyBitVectorPick(
        fps,
        poolSize=data_size,
        pickSize=test_size + valid_size,
        firstPicks=test_idxs,
        seed=random_state,
    )
    valid_idxs = list(set(valid_idxs) - set(test_idxs))

    train_idxs = list(set(range(data_size)) - set(test_idxs) - set(valid_idxs))

    ensure_nonempty_subset(train_idxs, "train")
    ensure_nonempty_subset(valid_idxs, "valid")
    ensure_nonempty_subset(test_idxs, "test")

    if return_indices:
        train_subset = train_idxs
        test_subset = test_idxs
        valid_subset = valid_idxs
    else:
        train_subset = get_data_from_indices(data, train_idxs)
        test_subset = get_data_from_indices(data, test_idxs)
        valid_subset = get_data_from_indices(data, valid_idxs)

    if additional_data:
        additional_data_split: list[Sequence[Any]] = split_additional_data(
            list(additional_data), train_idxs, valid_idxs, test_idxs
        )
        return train_subset, valid_subset, test_subset, additional_data_split
    else:
        return train_subset, valid_subset, test_subset



from .beyond_ro5 import BeyondRo5Filter
from .bms import BMSFilter
from .brenk import BrenkFilter
from .faf4_druglike import FAF4DruglikeFilter
from .faf4_leadlike import FAF4LeadlikeFilter
from .ghose import GhoseFilter
from .glaxo import GlaxoFilter
from .gsk import GSKFilter
from .hao import HaoFilter
from .inpharmatica import InpharmaticaFilter
from .lint import LINTFilter
from .lipinski import LipinskiFilter
from .mlsmr import MLSMRFilter
from .mol_weight import MolecularWeightFilter
from .nibr import NIBRFilter
from .nih import NIHFilter
from .oprea import OpreaFilter
from .pains import PAINSFilter
from .pfizer import PfizerFilter
from .reos import REOSFilter
from .rule_of_four import RuleOfFourFilter
from .rule_of_three import RuleOfThreeFilter
from .rule_of_two import RuleOfTwoFilter
from .rule_of_veber import RuleOfVeberFilter
from .rule_of_xu import RuleOfXuFilter
from .surechembl import SureChEMBLFilter
from .tice_herbicides import TiceHerbicidesFilter
from .tice_insecticides import TiceInsecticidesFilter
from .valence_discovery import ValenceDiscoveryFilter
from .zinc_basic import ZINCBasicFilter
from .zinc_druglike import ZINCDruglikeFilter



import warnings
from collections.abc import Sequence
from numbers import Integral
from typing import Optional, Union

import numpy as np
from rdkit.Chem import Mol, MolToSmiles
from scipy.sparse import csr_array
from sklearn.utils._param_validation import Interval, InvalidParameterError

from skfp.bases import BaseFingerprintTransformer
from skfp.utils import ensure_mols, require_mols_with_conf_ids


class AtomPairFingerprint(BaseFingerprintTransformer):
    """
    Atom Pair fingerprint.

    The implementation uses RDKit. This is a hashed fingerprint [1]_, where
    fragments are computed based on pairs of atoms and distance between them.

    Concretely, the hashed fragment is a triplet: (atom type 1, atom type 2, distance)

    Atom type takes into consideration:

    - atomic number
    - number of pi electrons
    - degree (number of bonds)

    Distance is normally the topological distance, i.e. length of the shortest path
    in the molecular graph (number of bonds between atoms). Only pairs with distance
    between ``min_distance`` and ``max_distance`` (both inclusive) are used.

    If ``use_3D`` is True, then the Euclidean distance between atoms in a conformation
    is used. Note that this uses ``conf_id`` property of input molecules, and requires
    them to have this property set.

    Values of count version are sensitive to the molecule size, since the number of
    shortest paths scales with square of heavy atom count (HAC). This can be offset
    by setting ``scale_by_hac`` to True (divide counts by HAC), or integer value greater
    than 1, which divides by HAC to the given power. Setting ``scale_by_hac=2`` makes
    valeus independent of molecule size.

    Parameters
    ----------
    fp_size : int, default=2048
        Size of output vectors, i.e. number of bits for each fingerprint. Must be
        positive.

    min_distance : int, default = 1
        Minimal distance between atoms. Must be positive and less or equal to
        ``max_distance``.

    max_distance : int, default = 30
        Maximal distance between atoms. Must be positive and greater or equal to
        ``min_distance``.

    use_pharmacophoric_invariants: bool, default=False
        Whether to use pharmacophoric invariants (atom types) instead of default ones.
        They are the same as in the FCFP fingerprint: Donor, Acceptor, Aromatic,
        Halogen, Basic, Acidic.

    include_chirality : bool, default=False
        Whether to include chirality information when computing atom types.

    use_3D : bool, default=False
        Whether to use 3D Euclidean distance matrix. If False, uses topological
        distances on molecular graph.

    count_simulation : bool, default=True
        Whether to use count simulation for approximating feature counts [3]_.

    count : bool, default=False
        Whether to return binary (bit) features, or their counts.

    scale_by_hac: bool or int, default=False
        Whether to scale count fingerprint by the heavy atom count (HAC) to
        obtain a proportionality to molecule size [2]_. If integer value is given,
        scaling uses given power of HAC, e.g. ``scale_by_hac=2`` divides counts by
        squared HAC. Using squared HAC results in values range [0, 1].

    sparse : bool, default=False
        Whether to return dense NumPy array, or sparse SciPy CSR array.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int
        Number of output features, size of fingerprints. Equal to ``fp_size``.

    requires_conformers : bool
        Whether the fingerprint is 3D-based and requires molecules with conformers as
        inputs, with ``conf_id`` integer property set. This depends on the ``use_3D``
        parameter, and has the same value.

    See Also
    --------
    :class:`TopologicalTorsionFingerprint` : Related fingerprint, but uses 4-atom paths.

    :class:`SECFPFingerprint` : Related fingerprint, which additionally uses circular
        substructures around each atom like ECFP fingerprint.

    References
    ----------
    .. [1] `Raymond E. Carhart, Dennis H. Smith, and R. Venkataraghavan
        "Atom pairs as molecular features in structure-activity studies: definition
        and applications"
        J. Chem. Inf. Comput. Sci. 1985, 25, 2, 64–73
        <https://pubs.acs.org/doi/10.1021/ci00046a002>`_

    .. [2] `Mahendra Awale and Jean-Louis Reymond
        "Atom Pair 2D-Fingerprints Perceive 3D-Molecular Shape and Pharmacophores for
        Very Fast Virtual Screening of ZINC and GDB-17"
        J. Chem. Inf. Model. 2014, 54, 7, 1892–1907
        <https://pubs.acs.org/doi/10.1021/ci500232g>`_

    .. [3] `Gregory Landrum
        "Simulating count fingerprints"
        <https://greglandrum.github.io/rdkit-blog/posts/2021-07-06-simulating-counts.html>`_

    Examples
    --------
    >>> from skfp.fingerprints import AtomPairFingerprint
    >>> smiles = ["O", "CC", "[C-]#N", "CC=O"]
    >>> fp = AtomPairFingerprint()
    >>> fp
    AtomPairFingerprint()

    >>> fp.transform(smiles)
    array([[0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0]], dtype=uint8)
    """

    _parameter_constraints: dict = {
        **BaseFingerprintTransformer._parameter_constraints,
        "fp_size": [Interval(Integral, 1, None, closed="left")],
        "min_distance": [Interval(Integral, 1, None, closed="left")],
        "max_distance": [Interval(Integral, 1, None, closed="left")],
        "include_chirality": ["boolean"],
        "use_pharmacophoric_invariants": ["boolean"],
        "use_3D": ["boolean"],
        "count_simulation": ["boolean"],
        "scale_by_hac": ["boolean", Interval(Integral, 0, None, closed="left")],
    }

    def __init__(
        self,
        fp_size: int = 2048,
        min_distance: int = 1,
        max_distance: int = 30,
        use_pharmacophoric_invariants: bool = False,
        include_chirality: bool = False,
        use_3D: bool = False,
        count_simulation: bool = True,
        scale_by_hac: Union[bool, int] = False,
        count: bool = False,
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            n_features_out=fp_size,
            requires_conformers=use_3D,
            count=count,
            sparse=sparse,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self.fp_size = fp_size
        self.min_distance = min_distance
        self.max_distance = max_distance
        self.use_pharmacophoric_invariants = use_pharmacophoric_invariants
        self.include_chirality = include_chirality
        self.use_3D = use_3D
        self.count_simulation = count_simulation
        self.scale_by_hac = scale_by_hac

    def _validate_params(self) -> None:
        super()._validate_params()
        if self.max_distance < self.min_distance:
            raise InvalidParameterError(
                f"The max_distance parameter of {self.__class__.__name__} must be "
                f"greater or equal to min_distance, got: "
                f"min_distance={self.min_distance}, max_distance={self.max_distance}"
            )

    def transform(
        self, X: Sequence[Union[str, Mol]], copy: bool = False
    ) -> Union[np.ndarray, csr_array]:
        """
        Compute Atom Pair fingerprints.

        Parameters
        ----------
        X : {sequence, array-like} of shape (n_samples,)
            Sequence containing SMILES strings or RDKit Mol objects. If ``use_3D``
            is True, only Mol objects with computed conformations and with
            ``conf_id`` property are allowed.

        copy : bool, default=False
            Copy the input X or not.

        Returns
        -------
        X : {ndarray, sparse matrix} of shape (n_samples, self.fp_size)
            Array with fingerprints.
        """
        return super().transform(X, copy)

    def _calculate_fingerprint(
        self, X: Sequence[Union[str, Mol]]
    ) -> Union[np.ndarray, csr_array]:
        from rdkit.Chem.rdFingerprintGenerator import (
            GetAtomPairGenerator,
            GetMorganFeatureAtomInvGen,
        )

        if self.use_3D:
            X = require_mols_with_conf_ids(X)
            conf_ids = [mol.GetIntProp("conf_id") for mol in X]
        else:
            X = ensure_mols(X)
            conf_ids = [-1 for _ in X]

        if self.use_pharmacophoric_invariants:
            inv_gen = GetMorganFeatureAtomInvGen()
        else:
            inv_gen = None

        gen = GetAtomPairGenerator(
            fpSize=self.fp_size,
            minDistance=self.min_distance,
            maxDistance=self.max_distance,
            atomInvariantsGenerator=inv_gen,
            includeChirality=self.include_chirality,
            countSimulation=self.count_simulation,
            use2D=not self.use_3D,
        )
        if self.count:
            fps = [
                gen.GetCountFingerprintAsNumPy(mol, confId=conf_id)
                for mol, conf_id in zip(X, conf_ids)
            ]
        else:
            fps = [
                gen.GetFingerprintAsNumPy(mol, confId=conf_id)
                for mol, conf_id in zip(X, conf_ids)
            ]

        if self.scale_by_hac:
            if self.count:
                fps = [self._scale_by_hac(fp, mol) for fp, mol in zip(fps, X)]
            else:
                warnings.warn(
                    "Scaling by HAC can only be applied to count vectors. "
                    "No HAC scaling will be applied."
                )

        return csr_array(fps) if self.sparse else np.array(fps)

    def _scale_by_hac(self, fingerprint: np.ndarray, mol: Mol) -> np.ndarray:
        hac = mol.GetNumHeavyAtoms()
        scale_factor = hac**self.scale_by_hac
        if scale_factor == 0:
            raise ZeroDivisionError(
                f"Got zero HAC scaling factor for SMILES '{MolToSmiles(mol)}' with "
                f"{hac} heavy atoms and scaling exponent {int(self.scale_by_hac)}"
            )

        return (100 * fingerprint) / scale_factor



from typing import Union

import numpy as np
from numba import njit, prange
from scipy.sparse import csr_array
from scipy.spatial.distance import dice
from sklearn.utils._param_validation import validate_params

from .utils import _check_nan


@validate_params(
    {
        "vec_a": ["array-like", csr_array],
        "vec_b": ["array-like", csr_array],
    },
    prefer_skip_nested_validation=True,
)
def dice_binary_similarity(
    vec_a: Union[np.ndarray, csr_array], vec_b: Union[np.ndarray, csr_array]
) -> float:
    """
    Dice similarity for vectors of binary values.

    Computes the Dice similarity [1]_ [2]_ [3]_ for binary data between two input arrays
    or sparse matrices using the formula:

    .. math::

        sim(vec_a, vec_b) = \\frac{2 |vec_a \\cap vec_b|}{|vec_a| + |vec_b|}

    The calculated similarity falls within the range ``[0, 1]``.
    Passing all-zero vectors to this function results in a similarity of 1.

    Parameters
    ----------
    vec_a : {ndarray, sparse matrix}
        First binary input array or sparse matrix.

    vec_b : {ndarray, sparse matrix}
        Second binary input array or sparse matrix.

    Returns
    ----------
    similarity : float
        Dice similarity between ``vec_a`` and ``vec_b``.

    References
    ----------
    .. [1] Dice, Lee R.
        "Measures of the amount of ecologic association between species."
        Ecology 26.3 (1945): 297-302.
        <https://esajournals.onlinelibrary.wiley.com/doi/abs/10.2307/1932409>

    .. [2] Brusco M., Cradit J. D., Steinley D.
        "A comparison of 71 binary similarity coefficients: The effect of base rates."
        PloS one 16.4 (2021): e0247751.
        <https://doi.org/10.1371/journal.pone.0247751>

    .. [3] Todeschini R., Consonni V., Xiang H., Holliday J., Buscema M., Willett P.
        "Similarity Coefficients for Binary Chemoinformatics Data: Overview and
        Extended Comparison Using Simulated and Real Data Sets."
        Journal of Chemical Information and Modeling 52.11 (2012): 2884-2901.
        <https://doi.org/10.1021/ci300261r>


    Examples
    ----------
    >>> from skfp.distances import dice_binary_similarity
    >>> import numpy as np
    >>> vec_a = np.array([1, 0, 1])
    >>> vec_b = np.array([1, 0, 1])
    >>> sim = dice_binary_similarity(vec_a, vec_b)
    >>> sim
    1.0

    >>> from skfp.distances import dice_binary_similarity
    >>> from scipy.sparse import csr_array
    >>> vec_a = csr_array([[1, 0, 1]])
    >>> vec_b = csr_array([[1, 0, 1]])
    >>> sim = dice_binary_similarity(vec_a, vec_b)
    >>> sim
    1.0
    """
    _check_nan(vec_a)
    _check_nan(vec_b)

    if np.sum(vec_a) == 0 == np.sum(vec_b):
        return 1.0

    if isinstance(vec_a, csr_array) and isinstance(vec_b, csr_array):
        vec_a_bool = vec_a.astype(bool)
        vec_b_bool = vec_b.astype(bool)
        intersection: float = vec_a_bool.multiply(vec_b_bool).sum()
        return 2 * intersection / (vec_a_bool.sum() + vec_b_bool.sum())
    elif isinstance(vec_a, np.ndarray) and isinstance(vec_b, np.ndarray):
        vec_a_bool = vec_a.astype(bool)
        vec_b_bool = vec_b.astype(bool)
        return 1 - dice(vec_a_bool, vec_b_bool)
    else:
        raise TypeError(
            f"Both vec_a and vec_b must be of the same type, either numpy.ndarray "
            f"or scipy.sparse.csr_array, got {type(vec_a)} and {type(vec_b)}"
        )


@validate_params(
    {
        "vec_a": ["array-like", csr_array],
        "vec_b": ["array-like", csr_array],
    },
    prefer_skip_nested_validation=True,
)
def dice_binary_distance(
    vec_a: Union[np.ndarray, csr_array], vec_b: Union[np.ndarray, csr_array]
) -> float:
    """
    Dice distance for vectors of binary values.

    Computes the Dice distance for binary data between two input arrays
    or sparse matrices by subtracting the Dice similarity [1]_ [2]_ [3]_ from 1,
    using to the formula:

    .. math::

        dist(vec_a, vec_b) = 1 - sim(vec_a, vec_b)

    The calculated distance falls within the range ``[0, 1]``.
    Passing all-zero vectors to this function results in a distance of 0.

    Parameters
    ----------
    vec_a : {ndarray, sparse matrix}
        First binary input array or sparse matrix.

    vec_b : {ndarray, sparse matrix}
        Second binary input array or sparse matrix.

    Returns
    ----------
    distance : float
        Dice distance between ``vec_a`` and ``vec_b``.

    References
    ----------
    .. [1] Dice, Lee R.
        "Measures of the amount of ecologic association between species."
        Ecology 26.3 (1945): 297-302.
        <https://esajournals.onlinelibrary.wiley.com/doi/abs/10.2307/1932409>

    .. [2] Brusco M., Cradit J. D., Steinley D.
        "A comparison of 71 binary similarity coefficients: The effect of base rates."
        PloS one 16.4 (2021): e0247751.
        <https://doi.org/10.1371/journal.pone.0247751>

    .. [3] Todeschini R., Consonni V., Xiang H., Holliday J., Buscema M., Willett P.
        "Similarity Coefficients for Binary Chemoinformatics Data: Overview and
        Extended Comparison Using Simulated and Real Data Sets."
        Journal of Chemical Information and Modeling 52.11 (2012): 2884-2901.
        <https://doi.org/10.1021/ci300261r>

    Examples
    ----------
    >>> from skfp.distances import dice_binary_distance
    >>> import numpy as np
    >>> vec_a = np.array([1, 0, 1])
    >>> vec_b = np.array([1, 0, 1])
    >>> dist = dice_binary_distance(vec_a, vec_b)
    >>> dist
    0.0

    >>> from skfp.distances import dice_binary_distance
    >>> from scipy.sparse import csr_array
    >>> vec_a = csr_array([[1, 0, 1]])
    >>> vec_b = csr_array([[1, 0, 1]])
    >>> dist = dice_binary_distance(vec_a, vec_b)
    >>> dist
    0.0
    """

    return 1 - dice_binary_similarity(vec_a, vec_b)


@validate_params(
    {
        "vec_a": ["array-like", csr_array],
        "vec_b": ["array-like", csr_array],
    },
    prefer_skip_nested_validation=True,
)
def dice_count_similarity(
    vec_a: Union[np.ndarray, csr_array], vec_b: Union[np.ndarray, csr_array]
) -> float:
    """
    Dice similarity for vectors of count values.

    Computes the Dice similarity [1]_ [2]_ [3]_ for count data between two input arrays
    or sparse matrices using the formula:

    .. math::

        sim(vec_a, vec_b) = \\frac{2 \\cdot vec_a \\cdot vec_b}{\\|vec_a\\|^2 + \\|vec_b\\|^2}

    The calculated similarity falls within the range ``[0, 1]``.
    Passing all-zero vectors to this function results in a similarity of 1.

    Parameters
    ----------
    vec_a : {ndarray, sparse matrix}
        First count input array or sparse matrix.

    vec_b : {ndarray, sparse matrix}
        Second count input array or sparse matrix.

    Returns
    -------
    similarity : float
        Dice similarity between ``vec_a`` and ``vec_b``.

    References
    ----------
    .. [1] Dice, Lee R.
        "Measures of the amount of ecologic association between species."
        Ecology 26.3 (1945): 297-302.
        <https://esajournals.onlinelibrary.wiley.com/doi/abs/10.2307/1932409>

    .. [2] Brusco M., Cradit J. D., Steinley D.
        "A comparison of 71 binary similarity coefficients: The effect of base rates."
        PloS one 16.4 (2021): e0247751.
        <https://doi.org/10.1371/journal.pone.0247751>

    .. [3] Todeschini R., Consonni V., Xiang H., Holliday J., Buscema M., Willett P.
        "Similarity Coefficients for Binary Chemoinformatics Data: Overview and
        Extended Comparison Using Simulated and Real Data Sets."
        Journal of Chemical Information and Modeling 52.11 (2012): 2884-2901.
        <https://doi.org/10.1021/ci300261r>

    Examples
    --------
    >>> from skfp.distances import dice_count_similarity
    >>> import numpy as np
    >>> vec_a = np.array([7, 1, 1])
    >>> vec_b = np.array([7, 1, 2])
    >>> sim = dice_count_similarity(vec_a, vec_b)
    >>> sim
    0.9904761904761905

    >>> from skfp.distances import dice_count_similarity
    >>> from scipy.sparse import csr_array
    >>> vec_a = csr_array([[7, 1, 1]])
    >>> vec_b = csr_array([[7, 1, 2]])
    >>> sim = dice_count_similarity(vec_a, vec_b)
    >>> sim
    0.9904761904761905
    """
    _check_nan(vec_a)
    _check_nan(vec_b)

    if np.sum(vec_a) == 0 == np.sum(vec_b):
        return 1.0

    if isinstance(vec_a, csr_array) and isinstance(vec_b, csr_array):
        return _dice_count_scipy(vec_a, vec_b)
    elif isinstance(vec_a, np.ndarray) and isinstance(vec_b, np.ndarray):
        return _dice_count_numpy(vec_a, vec_b)
    else:
        raise TypeError(
            f"Both vec_a and vec_b must be of the same type, either numpy.ndarray "
            f"or scipy.sparse.csr_array, got {type(vec_a)} and {type(vec_b)}"
        )


@validate_params(
    {
        "vec_a": ["array-like", csr_array],
        "vec_b": ["array-like", csr_array],
    },
    prefer_skip_nested_validation=True,
)
def dice_count_distance(
    vec_a: Union[np.ndarray, csr_array], vec_b: Union[np.ndarray, csr_array]
) -> float:
    """
    Dice distance for vectors of count values.

    Computes the Dice distance for count data between two input arrays
    or sparse matrices by subtracting the Dice similarity [1]_ [2]_ [3]_ from 1,
    using the formula:

    .. math::

        dist(vec_a, vec_b) = 1 - sim(vec_a, vec_b)

    The calculated distance falls within the range ``[0, 1]``.
    Passing all-zero vectors to this function results in a distance of 0.

    Parameters
    ----------
    vec_a : {ndarray, sparse matrix}
        First count input array or sparse matrix.

    vec_b : {ndarray, sparse matrix}
        Second count input array or sparse matrix.

    Returns
    -------
    distance : float
        Dice distance between ``vec_a`` and ``vec_b``.

    References
    ----------
    .. [1] Dice, Lee R.
        "Measures of the amount of ecologic association between species."
        Ecology 26.3 (1945): 297-302.
        <https://esajournals.onlinelibrary.wiley.com/doi/abs/10.2307/1932409>

    .. [2] Brusco M., Cradit J. D., Steinley D.
        "A comparison of 71 binary similarity coefficients: The effect of base rates."
        PloS one 16.4 (2021): e0247751.
        <https://doi.org/10.1371/journal.pone.0247751>

    .. [3] Todeschini R., Consonni V., Xiang H., Holliday J., Buscema M., Willett P.
        "Similarity Coefficients for Binary Chemoinformatics Data: Overview and
        Extended Comparison Using Simulated and Real Data Sets."
        Journal of Chemical Information and Modeling 52.11 (2012): 2884-2901.
        <https://doi.org/10.1021/ci300261r>

    Examples
    --------
    >>> from skfp.distances import dice_count_distance
    >>> import numpy as np
    >>> vec_a = np.array([7, 1, 1])
    >>> vec_b = np.array([7, 1, 2])
    >>> dist = dice_count_distance(vec_a, vec_b)
    >>> dist
    0.00952380952380949

    >>> from skfp.distances import dice_count_distance
    >>> from scipy.sparse import csr_array
    >>> vec_a = csr_array([[7, 1, 1]])
    >>> vec_b = csr_array([[7, 1, 2]])
    >>> dist = dice_count_distance(vec_a, vec_b)
    >>> dist
    0.00952380952380949
    """

    return 1 - dice_count_similarity(vec_a, vec_b)


@njit(parallel=True)
def _dice_count_numpy(vec_a: np.ndarray, vec_b: np.ndarray) -> float:
    vec_a = vec_a.astype(np.float64).ravel()
    vec_b = vec_b.astype(np.float64).ravel()

    dot_ab = 0.0
    dot_aa = 0.0
    dot_bb = 0.0

    for i in prange(vec_a.shape[0]):
        dot_ab += vec_a[i] * vec_b[i]
        dot_aa += vec_a[i] * vec_a[i]
        dot_bb += vec_b[i] * vec_b[i]

    return 2 * dot_ab / (dot_aa + dot_bb)


def _dice_count_scipy(vec_a: csr_array, vec_b: csr_array) -> float:
    dot_ab = vec_a.multiply(vec_b).sum()
    dot_aa = vec_a.multiply(vec_a).sum()
    dot_bb = vec_b.multiply(vec_b).sum()

    return 2 * dot_ab / (dot_aa + dot_bb)



from collections.abc import Sequence
from contextlib import nullcontext
from numbers import Integral
from typing import Optional, Union

from rdkit.Chem import Mol, SanitizeMol
from rdkit.Chem.MolStandardize.rdMolStandardize import (
    CleanupInPlace,
    FragmentParentInPlace,
)

from skfp.bases import BasePreprocessor
from skfp.utils import ensure_mols, no_rdkit_logs


class MolStandardizer(BasePreprocessor):
    """
    Performs common molecule standardization operations.

    Applies the following cleanup transformations to the inputs:
    - create RDKit Mol objects, if SMILES strings are passed
    - sanitize [1]_ (performs basic validity checks)
    - if ``largest_fragment_only``, select the largest fragment for further processing
    - remove hydrogens
    - disconnect metal atoms
    - normalize (transform functional groups to normal form)
    - reionize

    See [1]_ [2]_ [3]_ [4]_ [5]_ for details, rationale and alternatives. Note that
    there is no one-size-fits-all solution, and here we use pretty minimalistic, most
    common steps. This class by design does not allow much parametrization, and for
    custom purposes you should build the pipeline yourself.

    New molecules are always returned, and any set properties are not kept, as this
    should normally be the first step in a pipeline.

    Parameters
    ----------
    largest_fragment_only : bool, default=False
        Whether to select only the largest fragment from each molecule.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized over
        the input molecules. ``None`` means 1 unless in a :obj:`joblib.parallel_backend`
        context. ``-1`` means using all processors. See Scikit-learn documentation on
        ``n_jobs`` for more details.

    verbose : int, default=0
        Controls the verbosity when standardizing molecules. By default, all warnings are
        turned off.

    References
    ----------
    .. [1] `Gregory Landrum
        "The RDKit Book: Molecular Sanitization"
        <https://www.rdkit.org/docs/RDKit_Book.html#molecular-sanitization>`_

    .. [2] `RDKit mailing list thread on molecular sanitization
        <https://sourceforge.net/p/rdkit/mailman/rdkit-discuss/thread/CANjYGkS%3Dqz3NbmdhGAEvfOec0o%3DCL%3D43aex%3DT1RGmUDgSb63og%40mail.gmail.com/>`_

    .. [3] `Matt Swain
        "MolVS Standardization"
        <https://molvs.readthedocs.io/en/latest/guide/standardize.html>`_

    .. [4] `Gregory Landrum
        "RSC OpenScience workshop: Chemical structure validation and standardization with the RDKit"
        <https://github.com/greglandrum/RSC_OpenScience_Standardization_202104>`_

    .. [5] `Jean-Paul Ebejer
        "Standardizing a molecule using RDKit"
        <https://bitsilla.com/blog/2021/06/standardizing-a-molecule-using-rdkit/>`_

    Examples
    --------
    >>> from skfp.preprocessing import MolStandardizer
    >>> smiles = ["O", "CC", "[C-]#N", "CC=O"]
    >>> standardizer = MolStandardizer()
    >>> standardizer
    MolStandardizer()

    >>> standardizer.transform(smiles)  # doctest: +SKIP
        [<rdkit.Chem.rdchem.Mol>,
         <rdkit.Chem.rdchem.Mol>,
         <rdkit.Chem.rdchem.Mol>,
         <rdkit.Chem.rdchem.Mol>]
    """

    _parameter_constraints: dict = {
        "largest_fragment_only": ["boolean"],
        "n_jobs": [Integral, None],
        "verbose": ["verbose"],
    }

    def __init__(
        self,
        largest_fragment_only: bool = False,
        n_jobs: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__()
        self.largest_fragment_only = largest_fragment_only
        self.n_jobs = n_jobs
        self.verbose = verbose

    def transform(self, X: Sequence[Union[str, Mol]], copy: bool = True) -> list[Mol]:
        """
        Standardize molecule structures.

        Parameters
        ----------
        X : {sequence, array-like} of shape (n_samples,)
            Sequence containing SMILES strings or RDKit ``Mol`` objects.

        copy : bool, default=True
            Copy the input X or not. In contrast to most classes, input molecules
            are copied by default, since RDKit standardizes their structure in place.

        Returns
        -------
        X : list of shape (n_samples_conf_gen,)
            List with RDKit ``Mol`` objects.
        """
        return super().transform(X, copy)

    def _transform_batch(self, X: Sequence[Union[str, Mol]]) -> list[Mol]:
        # if SMILES is provided, MolFromSmiles also includes sanitization
        mols = ensure_mols(X)

        with nullcontext() if self.verbose else no_rdkit_logs():
            # we can safely modify in place here, thanks to default copy=True
            for mol in mols:
                # basic sanitization, if Mol objects were passed
                SanitizeMol(mol)

                # select the largest ("parent") fragment if needed
                if self.largest_fragment_only:
                    FragmentParentInPlace(mol)

                # remove Hs, disconnect metals, normalize functional groups, reionize
                CleanupInPlace(mol)

        return mols



from collections.abc import Sequence
from typing import Optional, Union

import numpy as np
from rdkit.Chem import Mol, MolToSmiles
from scipy.sparse import csr_array

from skfp.bases import BaseFingerprintTransformer
from skfp.utils import ensure_mols


class MACCSFingerprint(BaseFingerprintTransformer):
    """
    MACCS fingerprint.

    The implementation uses RDKit. This is a substructure fingerprint, based on
    publicly available MDL definitions, and refined by Gregory Landrum for RDKit [1]_.
    Note that full public definitions are not available, and packages differ [2]_.

    Results differ slightly from RDKit, because it adds an artificial 0th feature with
    all zeros, to simulate 1-based indexing [3]_. We use exactly 166 features, removing
    said constant feature. First feature is also always zero, since it means "ISOTOPE",
    not supported by RDKit. Consider removing them before to further processing, e.g.
    using ``VarianceThreshold``.

    Count variant is an original one. It counts substructures instead of only checking
    for their existence. It also has fewer features, because RDKit MACCS has separate
    features checking e.g. the number of oxygens. The ordering of features also differs,
    and there are no constant zero features.

    Parameters
    ----------
    count : bool, default=False
        Whether to return binary (bit) features, or the count-based variant.

    sparse : bool, default=False
        Whether to return dense NumPy array, or sparse SciPy CSR array.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int = 166 or 159.
        Number of output features, size of fingerprints. Equal to 166 for the bit
        variant, and 159 for count.

    requires_conformers : bool = False
        This fingerprint uses only 2D molecular graphs and does not require conformers.

    References
    ----------
    .. [1] RDKit MACCS implementation
        https://github.com/rdkit/rdkit/blob/3457c1eb60846ea821e4a319f3505933027d3cf8/rdkit/Chem/MACCSkeys.py

    .. [2] `Andrew Dalke
        "MACCS key 44"
        <http://www.dalkescientific.com/writings/diary/archive/2014/10/17/maccs_key_44.html>`_

    .. [3] RDKit MACCS 167 features discussion
        https://github.com/rdkit/rdkit/issues/1726

    Examples
    --------
    >>> from skfp.fingerprints import MACCSFingerprint
    >>> smiles = ["O", "CC", "[C-]#N", "CC=O"]
    >>> fp = MACCSFingerprint()
    >>> fp
    MACCSFingerprint()

    >>> fp.transform(smiles)
    array([[0, 0, 0, ..., 1, 0, 0],
           [0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 1, 0, 0]], dtype=uint8)
    """

    def __init__(
        self,
        count: bool = False,
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        n_features_out = 159 if count else 166
        super().__init__(
            n_features_out=n_features_out,
            count=count,
            sparse=sparse,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )

    def _calculate_fingerprint(
        self, X: Sequence[Union[str, Mol]]
    ) -> Union[np.ndarray, csr_array]:
        from rdkit.Chem.rdMolDescriptors import GetMACCSKeysFingerprint

        X = ensure_mols(X)

        if self.count:
            X = [self._get_maccs_patterns_counts(mol) for mol in X]
        else:
            X = [GetMACCSKeysFingerprint(mol) for mol in X]
            X = np.array(X)[:, 1:]  # remove constant zeros column

        dtype = np.uint32 if self.count else np.uint8
        return csr_array(X, dtype=dtype) if self.sparse else np.array(X, dtype=dtype)

    def _get_maccs_patterns_counts(self, mol: Mol) -> list[int]:
        # flake8: noqa: E501
        smarts_list = [
            None,  # fragments
            None,  # atomic num >103 - idx 0
            "[#7]",  # N
            "[#8]",  # O
            "[#9]",  # F
            "[#14]",  # Si
            "[#15]",  # P
            "[#16]",  # S
            "[#17]",  # Cl
            "[#35]",  # Br
            "[#53]",  # I
            "a",  # Aromatic
            "[F,Cl,Br,I]",  # X (HALOGEN)
            "[!+0]",  # CHARGE
            "[!#6;!#1;!H0]",  # QH
            "[!#1;!#6;!#7;!#8;!#9;!#14;!#15;!#16;!#17;!#35;!#53]",  # OTHER
            "[Li,Na,K,Rb,Cs,Fr]",  # Group IA (Alkali Metal)
            "[Cu,Zn,Ag,Cd,Au,Hg]",  # Group IB,IIB (Cu..)
            "[Be,Mg,Ca,Sr,Ba,Ra]",  # Group IIa (Alkaline earth)
            "[#5,#13,#31,#49,#81]",  # Group IIIA (B...)
            "[Sc,Ti,Y,Zr,Hf]",  # Group IIIB,IVB (Sc...)
            "[#32,#33,#34,#50,#51,#52,#82,#83,#84]",  # Group IVa,Va,VIa Rows 4-6
            "[V,Cr,Mn,Nb,Mo,Tc,Ta,W,Re]",  # Group VB,VIB,VIIB
            "[Fe,Co,Ni,Ru,Rh,Pd,Os,Ir,Pt]",  # Group VIII (Fe...)
            "[Ac,Th,Pa,U,Np,Pu,Am,Cm,Bk,Cf,Es,Fm,Md,No,Lr]",  # actinide
            "[La,Ce,Pr,Nd,Pm,Sm,Eu,Gd,Tb,Dy,Ho,Er,Tm,Yb,Lu]",  # Lanthanide
            "[R]",  # Ring
            None,  # Aromatic Ring
            "[!#6;R]",  # Heterocyclic atom (&...) Spec Incomplete
            "[!C;!c;R]",  # Heterocycle
            "[#16R]",  # S Heterocycle
            "[#8R]",  # O Heterocycle
            "[#7;R]",  # N Heterocycle
            "*1~*~*~1",  # 3-membered ring
            "*1~*~*~*~1",  # 4-membered ring
            "*1~*~*~*~*~1",  # 5-membered ring
            "*1~*~*~*~*~*~1",  # 6-membered ring
            "*1~*~*~*~*~*~*~1",  # 7-membered ring
            None,  # 8-membered ring or larger - idx 38
            "[#16]-[#16]",  # S-S
            "[#7]-[#8]",  # N-O
            "[#7]~[#16]",  # NS
            "[#16]-[#8]",  # S-O
            "[#7]~[#7]",  # NN
            "[#16]=[#8]",  # S=O
            "[#7]=[#8]",  # N=O
            "c:n",  # C%N
            "[!#6;!#1]~[#16]",  # QS
            "[#7]~[#8]",  # NO
            "[#16]=*",  # S=A
            "[#6]=[#7]",  # C=N
            "[NH2]",  # NH2
            "[!#6;!#1]~[#7]",  # QN
            "[#6]=[#6]",  # C=C
            "[!#6;!#1]~[#8]",  # QO
            "[!#6;!#1]~[F,Cl,Br,I]",  # QX
            "[#7]=*",  # N=A
            "[!#6;!#1]~[!#6;!#1]",  # QQ
            "[!#6;!#1]~[!#6;!#1]",  # QQ (&...)  Spec Incomplete
            "[#8]=*",  # O=A>1
            "[O;!H0]",  # OH
            "[CH3]",  # CH3 (&...) Spec Incomplete
            "[C;H3,H4]",  # CH3
            "[#7;!H0]",  # NH
            "[#6]=[#8]",  # C=O
            "[#6]-[#8]",  # C-O
            "[#6]-[#7]",  # C-N
            "[C;H3,H4]",  # CH3
            "[#6]#[#6]",  # CTC
            "[#6]#[#7]",  # CTN
            "[#6]~[#16]~[#8]",  # CSO
            "[#6]~[#16]~[#7]",  # CSN
            "[CH2]=*",  # CH2=A
            "[#6]=[#6]~[#7]",  # C=CN
            "[#16]~*~[#7]",  # SAN
            "[#8]~[#16]~[#8]",  # OSO
            "[!#6;!#1]~[#16]~[!#6;!#1]",  # QSQ
            "[#16]!:*:*",  # Snot%A%A
            "[!#6;!#1]~[!#6;!#1;!H0]",  # QQH
            "[!#6;!#1]~[#7]~[!#6;!#1]",  # QNQ
            "[#7]~*~[#7]",  # NAN
            "[#7]~[#6]~[#8]",  # NCO
            "[#7]~*~[#8]",  # NAO
            "[#8]~[#6]~[#8]",  # OCO
            "*!@[#8]!@*",  # A!O!A
            "*@*!@*@*",  # A$!A$A
            "*@*!@[#16]",  # A$A!S
            "[F,Cl,Br,I]!@*@*",  # X!A$A
            "*@*!@[#8]",  # A$A!O (&...) Spec Incomplete
            "*@*!@[#7]",  # A$A!N
            "[#7]!:*:*",  # Nnot%A%A
            "*@*!@[#8]",  # A$A!O
            "[#7]~[#6](~[#8])~[#7]",  # NC(O)N
            "[!#6;!#1]1~*~*~*~1",  # QAAA@1
            "[#8]~[#7](~[#6])~[#6]",  # ON(C)C
            "[#8]~[#6](~[#8])~[#8]",  # OC(O)O
            "[!#6;!#1]1~*~*~1",  # QAA@1
            "[#6]=[#6](~[!#6;!#1])~[!#6;!#1]",  # C=C(Q)Q
            "[#7]~[#6](~[#8])~[#8]",  # NC(O)O
            "[#7]~[#6](~[#7])~[#7]",  # NC(N)N
            "[#6]=;@[#6](@*)@*",  # C$=C($A)$A
            "[!#6;!#1]~[CH2]~[!#6;!#1]",  # QCH2Q
            "[#6]~[!#6;!#1](~[#6])(~[#6])~*",  # CQ(C)(C)A
            "[#8]~[#16](~[#8])~[#8]",  # OS(O)O
            "[!#6;!#1;!H0]~*~[!#6;!#1;!H0]",  # QHAQH
            "[#8]~[!#6;!#1](~[#8])(~[#8])",  # OQ(O)O
            "[#6]=[#6](~[#6])~[#6]",  # C=C(C)C
            "[!#6;!#1;!H0]~*~*~*~[!#6;!#1;!H0]",  # QHAAAQH
            "[!#6;!#1;!H0]~*~*~[!#6;!#1;!H0]",  # QHAAQH
            "[#8]~[#7](~[#8])~[#6]",  # ON(O)C
            "*~[#16](~*)~*",  # AS(A)A
            "[#6]~[#6](~[#6])(~[#6])~*",  # CC(C)(C)A
            "[!#6;!#1;!H0]~[!#6;!#1;!H0]",  # QHQH (&...) SPEC Incomplete
            "[#8]~*~*~[#8]",  # OAAO
            "[CH3]~*~[CH3]",  # CH3ACH3
            "*!@[#7]@*",  # A!N$A
            "[#6]=[#6](~*)~*",  # C=C(A)A
            "[#7]~*~*~[#7]",  # NAAN
            "[#7]~*~*~*~[#7]",  # NAAAN
            "[#16]~*(~*)~*",  # SA(A)A
            "*~[CH2]~[!#6;!#1;!H0]",  # ACH2QH
            "[!#6;!#1]1~*~*~*~*~1",  # QAAAA@1
            "[#6]~[#7](~[#6])~[#6]",  # CN(C)C
            "[C;H2,H3][!#6;!#1][C;H2,H3]",  # CH2QCH2
            "[#8]~*~*~*~[#8]",  # OAAAO
            # QHAACH2A
            "[$([!#6;!#1;!H0]~*~*~[CH2]~*),$([!#6;!#1;!H0;R]1@[R]@[R]@[CH2;R]1),$([!#6;!#1;!H0]~[R]1@[R]@[CH2;R]1)]",
            # QHAAACH2A
            "[$([!#6;!#1;!H0]~*~*~*~[CH2]~*),$([!#6;!#1;!H0;R]1@[R]@[R]@[R]@[CH2;R]1),$([!#6;!#1;!H0]~[R]1@[R]@[R]@[CH2;R]1),$([!#6;!#1;!H0]~*~[R]1@[R]@[CH2;R]1)]",
            "[#8]~[#6](~[#7])~[#6]",  # OC(N)C
            "[!#6;!#1]~[CH3]",  # QCH3
            "[#7]~*~*~[#8]",  # NAAO
            "[#7]~*~*~*~[#8]",  # NAAAO
            "[!#6;!#1]1~*~*~*~*~*~1",  # QAAAAA@1
            "*~[CH2]~[#7]",  # ACH2N
            "[!#6;!#1;!H0]~*~[CH2]~*",  # QHACH2A
            "*@*(@*)@*",  # A$A($A)$A
            "[!#6;!#1]~*(~[!#6;!#1])~[!#6;!#1]",  # QA(Q)Q
            "[F,Cl,Br,I]~*(~*)~*",  # XA(A)A
            "[CH3]~*~*~*~[CH2]~*",  # CH3AAACH2A
            "*~[CH2]~[#8]",  # ACH2O
            "[#7]~*~[CH2]~*",  # NACH2A
            "*~*(~*)(~*)~*",  # AA(A)(A)A
            "[#8]!:*:*",  # Onot%A%A
            "[CH3]~[CH2]~*",  # CH3CH2A
            "[CH3]~*~[CH2]~*",  # CH3ACH2A
            "[$([CH3]~*~*~[CH2]~*),$([CH3]~*1~*~[CH2]1)]",  # CH3AACH2A
            "[$(*~[CH2]~[CH2]~*),$(*1~[CH2]~[CH2]1)]",  # ACH2CH2A
            "*~[#7](~*)~*",  # AN(A)A
            "[$(*~[CH2]~*~*~*~[CH2]~*),$([R]1@[CH2;R]@[R]@[R]@[R]@[CH2;R]1),$(*~[CH2]~[R]1@[R]@[R]@[CH2;R]1),$(*~[CH2]~*~[R]1@[R]@[CH2;R]1)]",
            # ACH2AAACH2A
            "[$(*~[CH2]~*~*~[CH2]~*),$([R]1@[CH2]@[R]@[R]@[CH2;R]1),$(*~[CH2]~[R]1@[R]@[CH2;R]1)]",  # ACH2AACH2A
            "[#8]~*~[CH2]~*",  # OACH2A
            "[!#6;!#1]~[CH2]~*",  # QCH2A (&...) Spec Incomplete
            "*!:*:*!:*",  # Anot%A%Anot%A
            "[$(*~[CH2]~[CH2]~*),$([R]1@[CH2;R]@[CH2;R]1)]",  # ACH2CH2A
            "*~[!#6;!#1](~*)~*",  # AQ(A)A
            "*!@*@*!@*",  # A!A$A!A
            "[#8]~[#6](~[#6])~[#6]",  # OC(C)C
            "[!#6;!#1]~[CH2]~*",  # QCH2A
            "*!@[CH2]!@*",  # A!CH2!A
            "[#7]~*(~*)~*",  # NA(A)A
        ]
        counts = self._get_smarts_match_counts(mol, smarts_list)

        # here we fix positions that can't be easily written as SMARTS

        # number of fragments
        counts[0] = MolToSmiles(mol).count(".") + 1

        # atomic number over 103
        counts[1] = sum(atom.GetAtomicNum() > 103 for atom in mol.GetAtoms())

        # aromatic rings
        counts[28] = sum(
            all(mol.GetBondWithIdx(bond_idx).GetIsAromatic() for bond_idx in ring)
            for ring in mol.GetRingInfo().BondRings()
        )

        # 8-membered rings or larger
        counts[39] = sum(1 for ring in mol.GetRingInfo().BondRings() if len(ring) >= 8)

        return counts

    def _get_smarts_match_counts(
        self, mol: Mol, smarts_list: list[Optional[str]]
    ) -> list[int]:
        from rdkit.Chem import MolFromSmarts

        return [
            (
                len(mol.GetSubstructMatches(MolFromSmarts(smarts)))
                if smarts is not None
                else 0
            )
            for smarts in smarts_list
        ]



from typing import Union

import numpy as np
from numba import njit, prange
from scipy.sparse import csr_array
from scipy.spatial.distance import jaccard
from sklearn.utils._param_validation import validate_params

from .utils import _check_nan


@validate_params(
    {
        "vec_a": ["array-like", csr_array],
        "vec_b": ["array-like", csr_array],
    },
    prefer_skip_nested_validation=True,
)
def tanimoto_binary_similarity(
    vec_a: Union[np.ndarray, csr_array], vec_b: Union[np.ndarray, csr_array]
) -> float:
    """
    Tanimoto similarity for vectors of binary values.

    Computes the Tanimoto similarity [1]_ for binary data between two input arrays
    or sparse matrices using the Jaccard index using formula:

    .. math::

        sim(vec_a, vec_b) = \\frac{|vec_a \\cap vec_b|}{|vec_a| + |vec_b| - |vec_a \\cap vec_b|}

    The calculated similarity falls within the range ``[0, 1]``.
    Passing all-zero vectors to this function results in a similarity of 1.

    Parameters
    ----------
    vec_a : {ndarray, sparse matrix}
        First binary input array or sparse matrix.

    vec_b : {ndarray, sparse matrix}
        Second binary input array or sparse matrix.

    Returns
    ----------
    similarity : float
        Tanimoto similarity between vec_a and vec_b.

    References
    ----------
    .. [1] `Bajusz, D., Rácz, A. & Héberger, K.
       "Why is Tanimoto index an appropriate choice for fingerprint-based similarity calculations?"
       J Cheminform, 7, 20 (2015).
       <https://jcheminf.biomedcentral.com/articles/10.1186/s13321-015-0069-3>`_

    Examples
    ----------
    >>> from skfp.distances import tanimoto_binary_similarity
    >>> import numpy as np
    >>> vec_a = np.array([1, 0, 1])
    >>> vec_b = np.array([1, 0, 1])
    >>> sim = tanimoto_binary_similarity(vec_a, vec_b)
    >>> sim
    1.0

    >>> from skfp.distances import tanimoto_binary_similarity
    >>> from scipy.sparse import csr_array
    >>> vec_a = csr_array([[1, 0, 1]])
    >>> vec_b = csr_array([[1, 0, 1]])
    >>> sim = tanimoto_binary_similarity(vec_a, vec_b)
    >>> sim
    1.0
    """
    _check_nan(vec_a)
    _check_nan(vec_b)

    if np.sum(vec_a) == 0 == np.sum(vec_b):
        return 1.0

    if isinstance(vec_a, csr_array) and isinstance(vec_b, csr_array):
        vec_a_bool = vec_a.astype(bool)
        vec_b_bool = vec_b.astype(bool)

        intersection: float = vec_a_bool.multiply(vec_b_bool).sum()
        union: float = vec_a_bool.sum() + vec_b_bool.sum() - intersection

        return intersection / union

    elif isinstance(vec_a, np.ndarray) and isinstance(vec_b, np.ndarray):
        vec_a_bool = vec_a.astype(bool)
        vec_b_bool = vec_b.astype(bool)

        return 1 - jaccard(vec_a_bool, vec_b_bool)
    else:
        raise TypeError(
            f"Both vec_a and vec_b must be of the same type, either numpy.ndarray "
            f"or scipy.sparse.csr_array, got {type(vec_a)} and {type(vec_b)}"
        )


@validate_params(
    {
        "vec_a": ["array-like", csr_array],
        "vec_b": ["array-like", csr_array],
    },
    prefer_skip_nested_validation=True,
)
def tanimoto_binary_distance(
    vec_a: Union[np.ndarray, csr_array], vec_b: Union[np.ndarray, csr_array]
) -> float:
    """
    Tanimoto distance for vectors of binary values.

    Computes the Tanimoto distance for binary data between two input arrays
    or sparse matrices by subtracting the similarity from 1, using to
    the formula:

    .. math::

        dist(vec_a, vec_b) = 1 - sim(vec_a, vec_b)

    The calculated distance falls within the range ``[0, 1]``.
    Passing all-zero vectors to this function results in a distance of 0.

    Parameters
    ----------
    vec_a : {ndarray, sparse matrix}
        First binary input array or sparse matrix.

    vec_b : {ndarray, sparse matrix}
        Second binary input array or sparse matrix.

    Returns
    ----------
    distance : float
        Tanimoto distance between ``vec_a`` and ``vec_b``.

    Examples
    ----------
    >>> from skfp.distances import tanimoto_binary_distance
    >>> import numpy as np
    >>> vec_a = np.array([1, 0, 1])
    >>> vec_b = np.array([1, 0, 1])
    >>> dist = tanimoto_binary_distance(vec_a, vec_b)
    >>> dist
    0.0

    >>> from skfp.distances import tanimoto_binary_distance
    >>> from scipy.sparse import csr_array
    >>> vec_a = csr_array([[1, 0, 1]])
    >>> vec_b = csr_array([[1, 0, 1]])
    >>> dist = tanimoto_binary_distance(vec_a, vec_b)
    >>> dist
    0.0
    """

    return 1 - tanimoto_binary_similarity(vec_a, vec_b)


@validate_params(
    {
        "vec_a": ["array-like", csr_array],
        "vec_b": ["array-like", csr_array],
    },
    prefer_skip_nested_validation=True,
)
def tanimoto_count_similarity(
    vec_a: Union[np.ndarray, csr_array], vec_b: Union[np.ndarray, csr_array]
) -> float:
    """
    Tanimoto similarity for vectors of count values.

    Computes the Tanimoto similarity [1]_ for count data between two input arrays
    or sparse matrices using the formula:

    .. math::

        sim(vec_a, vec_b) = \\frac{vec_a \\cdot vec_b}{\\|vec_a\\|^2 + \\|vec_b\\|^2 - vec_a \\cdot vec_b}

    Calculated similarity falls within the range of `[0, 1]`.
    Passing all-zero vectors to this function results in similarity of 1.

    Note that Numpy version is optimized with Numba JIT compiler, resulting in significantly faster
    performance compared to SciPy sparse arrays. First usage may be slightly slower due to Numba compilation.

    Parameters
    ----------
    vec_a : {ndarray, sparse matrix}
        First count input array or sparse matrix.

    vec_b : {ndarray, sparse matrix}
        Second count input array or sparse matrix.

    Returns
    ----------
    similarity : float
        Tanimoto similarity between vec_a and vec_b.

    References
    ----------
    .. [1] `Bajusz, D., Rácz, A. & Héberger, K.
       "Why is Tanimoto index an appropriate choice for fingerprint-based similarity calculations?"
       J Cheminform, 7, 20 (2015).
       <https://jcheminf.biomedcentral.com/articles/10.1186/s13321-015-0069-3>`_

    Examples
    ----------
    >>> from skfp.distances import tanimoto_count_similarity
    >>> import numpy as np
    >>> vec_a = np.array([7, 1, 1])
    >>> vec_b = np.array([7, 1, 2])
    >>> sim = tanimoto_count_similarity(vec_a, vec_b)
    >>> sim
    0.9811320754716981

    >>> from skfp.distances import tanimoto_count_similarity
    >>> from scipy.sparse import csr_array
    >>> vec_a = csr_array([[7, 1, 1]])
    >>> vec_b = csr_array([[7, 1, 2]])
    >>> sim = tanimoto_count_similarity(vec_a, vec_b)
    >>> sim
    0.9811320754716981
    """
    _check_nan(vec_a)
    _check_nan(vec_b)

    if np.sum(vec_a) == 0 and np.sum(vec_b) == 0:
        return 1.0

    if isinstance(vec_a, csr_array) and isinstance(vec_b, csr_array):
        return _tanimoto_count_scipy(vec_a, vec_b)
    elif isinstance(vec_a, np.ndarray) and isinstance(vec_b, np.ndarray):
        return _tanimoto_count_numpy(vec_a, vec_b)
    else:
        raise TypeError(
            f"Both vec_a and vec_b must be of the same type, either numpy.ndarray "
            f"or scipy.sparse.csr_array, got {type(vec_a)} and {type(vec_b)}"
        )


@validate_params(
    {
        "vec_a": ["array-like", csr_array],
        "vec_b": ["array-like", csr_array],
    },
    prefer_skip_nested_validation=True,
)
def tanimoto_count_distance(
    vec_a: Union[np.ndarray, csr_array], vec_b: Union[np.ndarray, csr_array]
) -> float:
    """
    Tanimoto distance for vectors of count values.

    Computes the Tanimoto distance for binary data between two input arrays
    or sparse matrices by subtracting similarity value from 1, using to
    the formula:

    .. math::

            dist(vec_a, vec_b) = 1 - sim(vec_a, vec_b)

    Calculated distance falls within the range from `[0, 1]`.
    Passing all-zero vectors to this function results in distance of 0.

    Parameters
    ----------
    vec_a : {ndarray, sparse matrix}
        First count input array or sparse matrix.

    vec_b : {ndarray, sparse matrix}
        Second count input array or sparse matrix.

    Returns
    ----------
    distance : float
        Tanimoto distance between vec_a and vec_b.

    Examples
    ----------
    >>> from skfp.distances import tanimoto_count_distance
    >>> import numpy as np
    >>> vec_a = np.array([7, 1, 1])
    >>> vec_b = np.array([7, 1, 2])
    >>> dist = tanimoto_count_distance(vec_a, vec_b)
    >>> dist
    0.018867924528301883

    >>> from skfp.distances import tanimoto_count_distance
    >>> from scipy.sparse import csr_array
    >>> vec_a = csr_array([[7, 1, 1]])
    >>> vec_b = csr_array([[7, 1, 2]])
    >>> dist = tanimoto_count_distance(vec_a, vec_b)
    >>> dist
    0.018867924528301883
    """

    return 1 - tanimoto_count_similarity(vec_a, vec_b)


@njit(parallel=True)
def _tanimoto_count_numpy(vec_a: np.ndarray, vec_b: np.ndarray) -> float:
    vec_a = vec_a.astype(np.float64).ravel()
    vec_b = vec_b.astype(np.float64).ravel()

    dot_ab = 0.0
    dot_aa = 0.0
    dot_bb = 0.0

    for i in prange(vec_a.shape[0]):
        dot_ab += vec_a[i] * vec_b[i]
        dot_aa += vec_a[i] * vec_a[i]
        dot_bb += vec_b[i] * vec_b[i]

    return dot_ab / (dot_aa + dot_bb - dot_ab)


def _tanimoto_count_scipy(vec_a: csr_array, vec_b: csr_array) -> float:
    dot_ab: float = vec_a.multiply(vec_b).sum()
    dot_aa: float = vec_a.multiply(vec_a).sum()
    dot_bb: float = vec_b.multiply(vec_b).sum()

    return dot_ab / (dot_aa + dot_bb - dot_ab)



from typing import Union

from rdkit.Chem import AssignStereochemistry, GetFormalCharge, Mol
from rdkit.Chem.Crippen import MolLogP
from rdkit.Chem.Descriptors import MolWt
from rdkit.Chem.rdMolDescriptors import (
    CalcNumAtomStereoCenters,
    CalcNumHBA,
    CalcNumHBD,
    CalcNumHeteroatoms,
    CalcNumRings,
    CalcNumRotatableBonds,
    CalcTPSA,
)

from skfp.bases.base_filter import BaseFilter

from .utils import (
    get_max_ring_size,
    get_non_carbon_to_carbon_ratio,
    get_num_carbon_atoms,
    get_num_charged_functional_groups,
    get_num_rigid_bonds,
)


class FAF4LeadlikeFilter(BaseFilter):
    """
    FAFDrugs4 Lead-Like Soft filter.

    Designed as a part of FAFDrugs4 software [1]_ [2]_. Based on literature describing
    physico-chemical properties of lead drugs. Designed to keep starting point molecules
    that can be further optimized, i.e. relatively small, with low logP, and that can
    be "decorated" further to increase affinity and or selectivity without becoming
    very ADMET unfriendly).

    Basically a more restrictive variant of FAFDrugs4 Drug-Like Soft filter.

    Molecule must fulfill conditions:

    - molecular weight in range ``[150, 400]``
    - logP in range ``[-3, 4]``
    - HBA <= 7
    - HBD <= 4
    - TPSA <= 160
    - number of rotatable bonds <= 9
    - number of rigid bonds <= 30
    - number of rings <= 4
    - max ring size <= 18
    - number of carbons in range ``[3, 35]``
    - number of heteroatoms in range ``[1, 15]``
    - non-carbons to carbons ratio in range ``[0.1, 1.1]``
    - number of charged functional groups <= 4
    - total formal charge in range ``[-4, 4]``
    - number of stereocenters <= 2

    Note that the FAF4Drugs uses ChemAxon for determining functional groups. We use
    their publicly available CXSMARTS list of functional groups [3]_. Phosphine and
    sulfoxide patterns could not be parsed by RDKit, so we manually fixed them.

    Parameters
    ----------
    allow_one_violation : bool, default=False
        Whether to allow violating one of the rules for a molecule. This makes the
        filter less restrictive.

    return_indicators : bool, default=False
        Whether to return a binary vector with indicators which molecules pass the
        filter, instead of list of molecules.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform_x_y` and
        :meth:`transform` are parallelized over the input molecules. ``None`` means 1
        unless in a :obj:`joblib.parallel_backend` context. ``-1`` means using all
        processors. See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when filtering molecules.

    References
    -----------
    .. [1] `Details of physico-chemical property filters available in FAF-Drugs4
        <https://fafdrugs4.rpbs.univ-paris-diderot.fr/filters.html>`_

    .. [2] `D. Lagorce et al.
        "FAF-Drugs4: free ADME-tox filtering computations for chemical biology and
        early stages drug discovery"
        Bioinformatics, 33(22), 2017, 3658–3660
        <https://doi.org/10.1093/bioinformatics/btx491>`_

    .. [3] `ChemAxon documentation: Predefined Functional Groups and Named Molecule Groups
        <https://docs.chemaxon.com/display/docs/attachments/attachments_1829721_1_functionalgroups.cxsmi>`_

    Examples
    ----------
    >>> from skfp.filters import FAF4LeadlikeFilter
    >>> smiles = ["C", "CC(=O)Nc1ccc(O)cc1"]
    >>> filt = FAF4LeadlikeFilter()
    >>> filt
    FAF4LeadlikeFilter()

    >>> filtered_mols = filt.transform(smiles)
    >>> filtered_mols
    ['CC(=O)Nc1ccc(O)cc1']
    """

    def __init__(
        self,
        allow_one_violation: bool = False,
        return_indicators: bool = False,
        n_jobs: Union[int, None] = None,
        batch_size: Union[int, None] = None,
        verbose: int = 0,
    ):
        super().__init__(
            allow_one_violation, return_indicators, n_jobs, batch_size, verbose
        )

    def _apply_mol_filter(self, mol: Mol) -> bool:
        # explicit "copy" via SMILES due to RDKit bug
        # https://github.com/rdkit/rdkit/issues/7917
        from rdkit.Chem import MolFromSmiles, MolToSmiles

        mol = MolFromSmiles(MolToSmiles(mol))
        AssignStereochemistry(mol)

        rules = [
            150 <= MolWt(mol) <= 400,
            -3 <= MolLogP(mol) <= 4,
            CalcNumHBA(mol) <= 7,
            CalcNumHBD(mol) <= 4,
            CalcTPSA(mol) <= 160,
            CalcNumRotatableBonds(mol) <= 9,
            get_num_rigid_bonds(mol) <= 30,
            CalcNumRings(mol) <= 4,
            get_max_ring_size(mol) <= 18,
            3 <= get_num_carbon_atoms(mol) <= 35,
            1 <= CalcNumHeteroatoms(mol) <= 15,
            0.1 <= get_non_carbon_to_carbon_ratio(mol) <= 1.1,
            get_num_charged_functional_groups(mol) <= 4,
            -4 <= GetFormalCharge(mol) <= 4,
            CalcNumAtomStereoCenters(mol) <= 2,
        ]
        passed_rules = sum(rules)

        if self.allow_one_violation:
            return passed_rules >= len(rules) - 1
        else:
            return passed_rules == len(rules)



from collections.abc import Sequence
from numbers import Integral
from typing import Optional, Union

import numpy as np
from rdkit.Chem import Mol
from scipy.sparse import csr_array
from sklearn.utils._param_validation import Interval, InvalidParameterError, StrOptions

from skfp.bases import BaseFingerprintTransformer
from skfp.utils.validators import ensure_smiles


class MHFPFingerprint(BaseFingerprintTransformer):
    """
    MinHashed FingerPrint (MHFP).

    The implementation uses RDKit. This is a hashed fingerprint [1]_, where
    fragments are computed based on circular substructures around each atom.

    Subgraphs are created around each atom with increasing radius, starting
    with just an atom itself. It is then transformed into a canonical SMILES
    and hashed into an integer. In each iteration, it is increased by another
    atom (one "hop" on the graph). The resulting hashes are MinHashed. Depending
    on ``variant`` argument, either those values are returned, or they are further
    folded (with modulo) into a vector of size ``fp_size``.

    Additionally, the SMILES strings of the symmetrized smallest set of smallest
    rings (SSSR) are included by default, to incorporate ring information for
    small radii.

    Parameters
    ----------
    fp_size : int, default=2048
        Size of output vectors, i.e. number of bits for each fingerprint. Must be
        positive.

    radius : int, default=3
        Number of iterations performed, i.e. maximum radius of resulting subgraphs.
        Another common notation uses diameter, therefore ECFP4 has radius 2.

    min_radius : int, default=1
        Initial radius of subgraphs.

    sssr_rings : bool, default=True
        Whether to include the symmetrized smallest set of smallest rings (SSSR)
        in addition to circular subgraphs.

    isomeric_smiles : bool, default=False
        Whether to use isomeric SMILES, instead of just the canonical SMILES.

    kekulize : bool, default=True
        Whether to kekulize the subgraphs before SMILES generation.

    variant : {"bit", "count", "raw_hashes"}, default="bit"
        Fingerprint variant. "raw_hashes" follows the original paper and results in
        raw integer values of hashes. "bit" and "count" result in integer vectors
        with hashes folded with modulo operation into a ``fp_size`` length.

    sparse : bool, default=False
        Whether to return dense NumPy array, or sparse SciPy CSR array.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int
        Number of output features. Equal to ``fp_size``.

    requires_conformers : bool = False
        This fingerprint uses only 2D molecular graphs and does not require conformers.

    See Also
    --------
    :class:`ECFPFingerprint` : Related fingerprint, which uses atom invariants instead
        of raw SMILES strings.

    :class:`SECFPFingerprint` : Related fingerprint, which hashes subgraph SMILES and
        folds the resulting vector.

    References
    ----------
    .. [1] `Daniel Probst and Jean-Louis Reymond
        "A probabilistic molecular fingerprint for big data settings"
        Journal of Cheminformatics 10, 1-12 (2018)
        <https://jcheminf.biomedcentral.com/articles/10.1186/s13321-018-0321-8>`_

    Examples
    --------
    >>> from skfp.fingerprints import MHFPFingerprint
    >>> smiles = ["O", "CC", "[C-]#N", "CC=O"]
    >>> fp = MHFPFingerprint()
    >>> fp
    MHFPFingerprint()

    >>> fp.transform(smiles)
    array([[0, 0, 0, ..., 0, 0, 1],
           [0, 1, 0, ..., 0, 0, 1],
           [1, 1, 0, ..., 1, 1, 0],
           [0, 0, 1, ..., 1, 0, 1]], dtype=uint8)
    """

    _parameter_constraints: dict = {
        **BaseFingerprintTransformer._parameter_constraints,
        "fp_size": [Interval(Integral, 1, None, closed="left")],
        "radius": [Interval(Integral, 0, None, closed="left")],
        "min_radius": [Interval(Integral, 0, None, closed="left")],
        "rings": ["boolean"],
        "isomeric": ["boolean"],
        "kekulize": ["boolean"],
        "variant": [StrOptions({"bit", "count", "raw_hashes"})],
    }

    def __init__(
        self,
        fp_size: int = 2048,
        radius: int = 3,
        min_radius: int = 1,
        sssr_rings: bool = True,
        isomeric_smiles: bool = False,
        kekulize: bool = True,
        variant: str = "bit",
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            n_features_out=fp_size,
            sparse=sparse,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self.fp_size = fp_size
        self.radius = radius
        self.min_radius = min_radius
        self.sssr_rings = sssr_rings
        self.isomeric_smiles = isomeric_smiles
        self.kekulize = kekulize
        self.variant = variant

    def _validate_params(self) -> None:
        super()._validate_params()
        if self.radius < self.min_radius:
            raise InvalidParameterError(
                f"The radius parameter of {self.__class__.__name__} must be "
                f"greater or equal to min_radius, got: "
                f"min_radius={self.min_radius}, radius={self.radius}"
            )

    def _calculate_fingerprint(
        self, X: Sequence[Union[str, Mol]]
    ) -> Union[np.ndarray, csr_array]:
        from rdkit.Chem.rdMHFPFingerprint import MHFPEncoder

        X = ensure_smiles(X)

        # outputs raw hash values, not feature vectors!
        encoder = MHFPEncoder(self.fp_size, self.random_state)
        X = MHFPEncoder.EncodeSmilesBulk(
            encoder,
            X,
            radius=self.radius,
            min_radius=self.min_radius,
            rings=self.sssr_rings,
            isomeric=self.isomeric_smiles,
            kekulize=self.kekulize,
        )
        X = np.array(X, dtype=np.uint32)

        if self.variant in {"bit", "count"}:
            X = np.mod(X, self.fp_size)
            X = np.stack([np.bincount(fp, minlength=self.fp_size) for fp in X])
            if self.variant == "bit":
                X = (X > 0).astype(np.uint8)
            else:
                X = X.astype(np.uint32)

        return csr_array(X) if self.sparse else np.array(X)



from .base_filter import BaseFilter
from .base_fp_transformer import BaseFingerprintTransformer
from .base_preprocessor import BasePreprocessor
from .base_substructure_fp import BaseSubstructureFingerprint



from collections.abc import Sequence
from typing import Optional, Union

import numpy as np
from rdkit.Chem import Mol
from scipy.sparse import csr_array

from skfp.bases import BaseSubstructureFingerprint


class GhoseCrippenFingerprint(BaseSubstructureFingerprint):
    """
    Ghose-Crippen fingerprint.

    A substructure fingerprint based on 110 atom types proposed by Ghose and
    Crippen [1]_ [2]_. They are defined for carbon, hydrogen, oxygen, nitrogen, sulfur,
    and halogens, and originally applied for predicting molar refractivities and logP.

    RDKit SMARTS patterns definitions are used [3]_.

    Parameters
    ----------
    count : bool, default=False
        Whether to return binary (bit) features, or their counts.

    sparse : bool, default=False
        Whether to return dense NumPy array, or sparse SciPy CSR array.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int = 110
        Number of output features, size of fingerprints.

    requires_conformers : bool = False
        This fingerprint uses only 2D molecular graphs and does not require conformers.

    References
    ----------
    .. [1] `Arup K. Ghose and Gordon M. Crippen
        "Atomic Physicochemical Parameters for Three-Dimensional Structure-Directed Quantitative Structure-Activity Relationships I. Partition Coefficients as a Measure of Hydrophobicity"
        Journal of Computational Chemistry 7.4 (1986): 565-577.
        <https://onlinelibrary.wiley.com/doi/abs/10.1002/jcc.540070419>`_

    .. [2] `Arup K. Ghose and Gordon M. Crippen
        "Atomic physicochemical parameters for three-dimensional-structure-directed quantitative structure-activity relationships. 2. Modeling dispersive and hydrophobic interactions"
        J. Chem. Inf. Comput. Sci. 1987, 27, 1, 21–35
        <https://pubs.acs.org/doi/10.1021/ci00053a005>`_

    .. [3] `<https://github.com/rdkit/rdkit/blob/5d034e37331c2604bf3e247b94be35b519e62216/Data/Crippen.txt>`_

    Examples
    --------
    >>> from skfp.fingerprints import GhoseCrippenFingerprint
    >>> smiles = ["O", "CC", "[C-]#N", "CC=O"]
    >>> fp = GhoseCrippenFingerprint()
    >>> fp
    GhoseCrippenFingerprint()

    >>> fp.transform(smiles)  # doctest: +ELLIPSIS
    array([[0, 0, 0, ..., 0, 0, 0],
           [0, 1, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0],
           [0, 1, 0, ..., 0, 0, 0]],
          dtype=uint8)
    """

    def __init__(
        self,
        count: bool = False,
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        # flake8: noqa: E501
        patterns = [
            "[CH4]",
            "[CH3]C",
            "[CH2](C)C",
            "[CH](C)(C)C",
            "[C](C)(C)(C)C",
            "[CH3][N,O,P,S,F,Cl,Br,I]",
            "[CH2X4]([N,O,P,S,F,Cl,Br,I])[A;!#1]",
            "[CH1X4]([N,O,P,S,F,Cl,Br,I])([A;!#1])[A;!#1]",
            "[CH0X4]([N,O,P,S,F,Cl,Br,I])([A;!#1])([A;!#1])[A;!#1]",
            "[C]=[!C;A;!#1]",
            "[CH2]=C",
            "[CH1](=C)[A;!#1]",
            "[CH0](=C)([A;!#1])[A;!#1]",
            "[C](=C)=C",
            "[CX2]#[A;!#1]",
            "[CH3]c",
            "[CH3]a",
            "[CH2X4]a",
            "[CHX4]a",
            "[CH0X4]a",
            "[cH0]-[A;!C;!N;!O;!S;!F;!Cl;!Br;!I;!#1]",
            "[c][#9]",
            "[c][#17]",
            "[c][#35]",
            "[c][#53]",
            "[cH]",
            "[c](:a)(:a):a",
            "[c](:a)(:a)-a",
            "[c](:a)(:a)-C",
            "[c](:a)(:a)-N",
            "[c](:a)(:a)-O",
            "[c](:a)(:a)-S",
            "[c](:a)(:a)=[C,N,O]",
            "[C](=C)(a)[A;!#1]",
            "[C](=C)(c)a",
            "[CH1](=C)a",
            "[C]=c",
            "[CX4][A;!C;!N;!O;!P;!S;!F;!Cl;!Br;!I;!#1]",
            "[#6]",
            "[#1][#6,#1]",
            "[#1]O[CX4,c]",
            "[#1]O[!C;!N;!O;!S]",
            "[#1][!C;!N;!O]",
            "[#1][#7]",
            "[#1]O[#7]",
            "[#1]OC=[#6,#7,O,S]",
            "[#1]O[O,S]",
            "[#1]",
            "[NH2+0][A;!#1]",
            "[NH+0]([A;!#1])[A;!#1]",
            "[NH2+0]a",
            "[NH1+0]([!#1;A,a])a",
            "[NH+0]=[!#1;A,a]",
            "[N+0](=[!#1;A,a])[!#1;A,a]",
            "[N+0]([A;!#1])([A;!#1])[A;!#1]",
            "[N+0](a)([!#1;A,a])[A;!#1]",
            "[N+0](a)(a)a",
            "[N+0]#[A;!#1]",
            "[NH3,NH2,NH;+,+2,+3]",
            "[n+0]",
            "[n;+,+2,+3]",
            "[NH0;+,+2,+3]([A;!#1])([A;!#1])([A;!#1])[A;!#1]",
            "[NH0;+,+2,+3](=[A;!#1])([A;!#1])[!#1;A,a]",
            "[NH0;+,+2,+3](=[#6])=[#7]",
            "[N;+,+2,+3]#[A;!#1]",
            "[N;-,-2,-3]",
            "[N;+,+2,+3](=[N;-,-2,-3])=N",
            "[#7]",
            "[o]",
            "[OH,OH2]",
            "[O]([A;!#1])[A;!#1]",
            "[O](a)[!#1;A,a]",
            "[O]=[#7,#8]",
            "[OX1;-,-2,-3][#7]",
            "[OX1;-,-2,-2][#16]",
            "[O;-0]=[#16;-0]",
            "[O-]C(=O)",
            "[OX1;-,-2,-3][!#1;!N;!S]",
            "[O]=c",
            "[O]=[CH]C",
            "[O]=C(C)([A;!#1])",
            "[O]=[CH][N,O]",
            "[O]=[CH2]",
            "[O]=[CX2]=O",
            "[O]=[CH]c",
            "[O]=C([C,c])[a;!#1]",
            "[O]=C(c)[A;!#1]",
            "[O]=C([!#1;!#6])[!#1;!#6]",
            "[#8]",
            "[#9-0]",
            "[#17-0]",
            "[#35-0]",
            "[#53-0]",
            "[#9,#17,#35,#53;-]",
            "[#53;+,+2,+3]",
            "[+;#3,#11,#19,#37,#55]",
            "[#15]",
            "[S;-,-2,-3,-4,+1,+2,+3,+5,+6]",
            "[S-0]=[N,O,P,S]",
            "[S;A]",
            "[s;a]",
            "[#3,#11,#19,#37,#55]",
            "[#4,#12,#20,#38,#56]",
            "[#5,#13,#31,#49,#81]",
            "[#14,#32,#50,#82]",
            "[#33,#51,#83]",
            "[#34,#52,#84]",
            "[#21,#22,#23,#24,#25,#26,#27,#28,#29,#30]",
            "[#39,#40,#41,#42,#43,#44,#45,#46,#47,#48]",
            "[#72,#73,#74,#75,#76,#77,#78,#79,#80]",
        ]
        # flake8: noqa
        super().__init__(
            patterns=patterns,
            count=count,
            sparse=sparse,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )

    def transform(
        self, X: Sequence[Union[str, Mol]], copy: bool = False
    ) -> Union[np.ndarray, csr_array]:
        """
        Compute CDK substructure fingerprints.

        Parameters
        ----------
        X : {sequence, array-like} of shape (n_samples,)
            Sequence containing SMILES strings or RDKit Mol objects.

        copy : bool, default=False
            Copy the input X or not.

        Returns
        -------
        X : {ndarray, sparse matrix} of shape (n_samples, 307)
            Array with fingerprints.
        """
        return super().transform(X, copy)



"""Classes for preprocessing molecules."""

from .conformer_generator import ConformerGenerator
from .input_output import (
    MolFromAminoseqTransformer,
    MolFromInchiTransformer,
    MolFromSDFTransformer,
    MolFromSmilesTransformer,
    MolToInchiTransformer,
    MolToSDFTransformer,
    MolToSmilesTransformer,
)
from .standardization import MolStandardizer



import os
from typing import Optional, Union

import numpy as np
import pandas as pd
from sklearn.utils._param_validation import validate_params

from skfp.datasets.utils import fetch_dataset, get_mol_strings_and_labels


@validate_params(
    {
        "data_dir": [None, str, os.PathLike],
        "as_frame": ["boolean"],
        "verbose": ["boolean"],
    },
    prefer_skip_nested_validation=True,
)
def load_clintox(
    data_dir: Optional[Union[str, os.PathLike]] = None,
    as_frame: bool = False,
    verbose: bool = False,
) -> Union[pd.DataFrame, tuple[list[str]], np.ndarray]:
    """
    Load and return the ClinTox dataset.

    The task is to predict drug approval viability, by predicting clinical trial
    toxicity and final FDA approval status [1]_. Both tasks are binary.

    ==================   ========================
    Tasks                                       2
    Task type            multitask classification
    Total samples                            1477
    Recommended split                    scaffold
    Recommended metric                      AUROC
    ==================   ========================

    Parameters
    ----------
    data_dir : {None, str, path-like}, default=None
        Path to the root data directory. If ``None``, currently set scikit-learn directory
        is used, by default `$HOME/scikit_learn_data`.

    as_frame : bool, default=False
        If True, returns the raw DataFrame with columns "SMILES" and 2 label columns,
        FDA approval and clinical trial toxicity. Otherwise, returns SMILES as list
        of strings, and labels as a NumPy array (2D integer array).

    verbose : bool, default=False
        If True, progress bar will be shown for downloading or loading files.

    Returns
    -------
    data : pd.DataFrame or tuple(list[str], np.ndarray)
        Depending on the ``as_frame`` argument, one of:
        - Pandas DataFrame with columns "SMILES" and 2 label columns
        - tuple of: list of strings (SMILES), NumPy array (labels)

    References
    ----------
    .. [1] `Zhenqin Wu et al.
        "MoleculeNet: a benchmark for molecular machine learning"
        Chem. Sci., 2018,9, 513-530
        <https://pubs.rsc.org/en/content/articlelanding/2018/sc/c7sc02664a>`_
    """
    df = fetch_dataset(
        data_dir,
        dataset_name="MoleculeNet_ClinTox",
        filename="clintox.csv",
        verbose=verbose,
    )
    return df if as_frame else get_mol_strings_and_labels(df)



from collections import defaultdict
from collections.abc import Sequence
from numbers import Integral
from typing import Any, Optional, Union

import numpy as np
from rdkit.Chem import Mol
from rdkit.Chem.rdFingerprintGenerator import GetMorganGenerator
from rdkit.SimDivFilters.rdSimDivPickers import LeaderPicker
from sklearn.neighbors import NearestNeighbors
from sklearn.utils._param_validation import Interval, RealNotInt, validate_params

from skfp.fingerprints import ECFPFingerprint
from skfp.model_selection.splitters.utils import (
    ensure_nonempty_subset,
    split_additional_data,
    validate_train_test_split_sizes,
    validate_train_valid_test_split_sizes,
)
from skfp.utils.functions import get_data_from_indices
from skfp.utils.validators import ensure_mols


@validate_params(
    {
        "data": ["array-like"],
        "additional_data": ["tuple"],
        "train_size": [
            Interval(RealNotInt, 0, 1, closed="neither"),
            Interval(Integral, 1, None, closed="left"),
            None,
        ],
        "test_size": [
            Interval(RealNotInt, 0, 1, closed="neither"),
            Interval(Integral, 1, None, closed="left"),
            None,
        ],
        "threshold": [Interval(RealNotInt, 0, 1, closed="both")],
        "approximate": ["boolean"],
        "return_indices": ["boolean"],
        "n_jobs": [Integral, None],
    },
    prefer_skip_nested_validation=True,
)
def butina_train_test_split(
    data: Sequence[Union[str, Mol]],
    *additional_data: Sequence,
    train_size: Optional[float] = None,
    test_size: Optional[float] = None,
    threshold: float = 0.65,
    approximate: bool = False,
    return_indices: bool = False,
    n_jobs: Optional[int] = None,
) -> Union[
    tuple[
        Sequence[Union[str, Mol]], Sequence[Union[str, Mol]], Sequence[Sequence[Any]]
    ],
    tuple[Sequence, ...],
    tuple[Sequence[int], Sequence[int]],
]:
    """
    Split using Taylor-Butina clustering.

    This split uses deterministically partitioned clusters of molecules from Taylor-Butina
    clustering [1]_ [2]_ [3]_. It aims to verify the model generalization to structurally
    novel molecules. Also known as sphere exclusion or leader-following clustering.

    First, molecules are vectorized using binary ECFP4 fingerprint (radius 2) with
    2048 bits. They are then clustered using Leader Clustering, a variant of Taylor-Butina
    clustering by Roger Sayle [4]_ for RDKit. Cluster centroids (central molecules) are
    guaranteed to have at least a given Tanimoto distance between them, as defined by
    `threshold` parameter.

    Clusters are divided deterministically, with the smallest clusters assigned to the
    test subset and the rest to the training subset.

    If ``train_size`` and ``test_size`` are integers, they must sum up to the ``data``
    length. If they are floating numbers, they must sum up to 1.

    Parameters
    ----------
    data : sequence
        A sequence representing either SMILES strings or RDKit ``Mol`` objects.

    additional_data: list[sequence]
        Additional sequences to be split alongside the main data (e.g., labels or feature vectors).

    train_size : float, default=None
        The fraction of data to be used for the train subset. If None, it is set to 1 - test_size.
        If test_size is also None, it will be set to 0.8.

    test_size : float, default=None
        The fraction of data to be used for the test subset. If None, it is set to 1 - train_size.
        If train_size is also None, it will be set to 0.2.

    threshold : float, default=0.65
        Tanimoto distance threshold, defining the minimal distance between cluster
        centroids. Default value is based on ECFP4 activity threshold as determined
        by Roger Sayle [4]_.

    approximate : bool, default=False
        Whether to use approximate similarity calculation to speed up computation on
        large datasets. It uses NNDescent algorithm [5]_ [6]_ and requires `PyNNDescent`
        library to be installed. However, it is much slower on small datasets, and
        exact version is always used for data with less than 5000 molecules.

    return_indices : bool, default=False
        Whether the method should return the input object subsets, i.e. SMILES strings
        or RDKit ``Mol`` objects, or only the indices of the subsets instead of the data.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    Returns
    ----------
    subsets : tuple[list, list, ...]
        Tuple with train-test subsets of provided arrays. First two are lists of SMILES
        strings or RDKit ``Mol`` objects, depending on the input type. If `return_indices`
        is True, lists of indices are returned instead of actual data.

    References
    ----------
    .. [1] `Darko Butina
        "Unsupervised Data Base Clustering Based on Daylight's Fingerprint and Tanimoto
        Similarity: A Fast and Automated Way To Cluster Small and Large Data Sets"
        . Chem. Inf. Comput. Sci. 1999, 39, 4, 747–750
        <https://pubs.acs.org/doi/abs/10.1021/ci9803381>`_

    .. [2] `Robin Taylor
        "Simulation Analysis of Experimental Design Strategies for Screening Random
        Compounds as Potential New Drugs and Agrochemicals"
        J. Chem. Inf. Comput. Sci. 1995, 35, 1, 59–67
        <https://pubs.acs.org/doi/10.1021/ci00023a009>`_

    .. [3] `Noel O'Boyle
        "Taylor-Butina Clustering"
        <https://noel.redbrick.dcu.ie/R_clustering.html>`_

    .. [4] `Roger Sayle
        "2D similarity, diversity and clustering in RDKit"
        RDKit UGM 2019
        <https://www.nextmovesoftware.com/talks/Sayle_2DSimilarityDiversityAndClusteringInRdkit_RDKITUGM_201909.pdf>`_

    .. [5] `W. Dong et al.
        "Efficient k-nearest neighbor graph construction for generic similarity measures"
        Proceedings of the 20th International World Wide Web Conference (WWW '11).
        Association for Computing Machinery, New York, NY, USA, 577–586
        <https://doi.org/10.1145/1963405.1963487>`_

    .. [6] `Leland McInnes
        "PyNNDescent for fast Approximate Nearest Neighbors"
        <https://pynndescent.readthedocs.io/en/latest/>`_
    """
    train_size, test_size = validate_train_test_split_sizes(
        train_size, test_size, len(data)
    )

    clusters = _create_clusters(data, threshold, approximate, n_jobs)
    clusters.sort(key=len)

    train_idxs: list[int] = []
    test_idxs: list[int] = []

    for cluster in clusters:
        if len(test_idxs) < test_size:
            test_idxs.extend(cluster)
        else:
            train_idxs.extend(cluster)

    ensure_nonempty_subset(train_idxs, "train")
    ensure_nonempty_subset(test_idxs, "test")

    if return_indices:
        train_subset = train_idxs
        test_subset = test_idxs
    else:
        train_subset = get_data_from_indices(data, train_idxs)
        test_subset = get_data_from_indices(data, test_idxs)

    if additional_data:
        additional_data_split: list[Sequence[Any]] = split_additional_data(
            list(additional_data), train_idxs, test_idxs
        )
        return train_subset, test_subset, *additional_data_split
    else:
        return train_subset, test_subset


@validate_params(
    {
        "data": ["array-like"],
        "additional_data": ["tuple"],
        "train_size": [
            Interval(RealNotInt, 0, 1, closed="neither"),
            Interval(Integral, 1, None, closed="left"),
            None,
        ],
        "valid_size": [
            Interval(RealNotInt, 0, 1, closed="neither"),
            Interval(Integral, 1, None, closed="left"),
            None,
        ],
        "test_size": [
            Interval(RealNotInt, 0, 1, closed="neither"),
            Interval(Integral, 1, None, closed="left"),
            None,
        ],
        "threshold": [Interval(RealNotInt, 0, 1, closed="both")],
        "approximate": ["boolean"],
        "return_indices": ["boolean"],
        "n_jobs": [Integral, None],
    },
    prefer_skip_nested_validation=True,
)
def butina_train_valid_test_split(
    data: Sequence[Union[str, Mol]],
    *additional_data: Sequence,
    train_size: Optional[float] = None,
    valid_size: Optional[float] = None,
    test_size: Optional[float] = None,
    threshold: float = 0.65,
    approximate: bool = False,
    return_indices: bool = False,
    n_jobs: Optional[int] = None,
) -> Union[
    tuple[
        Sequence[Union[str, Mol]],
        Sequence[Union[str, Mol]],
        Sequence[Union[str, Mol]],
        Sequence[Sequence[Any]],
    ],
    tuple[Sequence, ...],
    tuple[Sequence[int], Sequence[int], Sequence[int]],
]:
    """
    Split using Taylor-Butina clustering.

    This split uses deterministically partitioned clusters of molecules from Taylor-Butina
    clustering [1]_ [2]_ [3]_. It aims to verify the model generalization to structurally
    novel molecules.

    First, molecules are vectorized using binary ECFP4 fingerprint (radius 2) with
    2048 bits. They are then clustered using Leader Clustering, a variant of Taylor-Butina
    clustering by Roger Sayle for RDKit [4]_. Cluster centroids (central molecules) are
    guaranteed to have at least a given Tanimoto distance between them, as defined by
    `threshold` parameter.

    Clusters are divided deterministically, with the smallest clusters assigned to the
    test subset, larger to the validation subset, and the rest to the training subset

    If ``train_size``, ``valid_size`` and ``test_size`` are integers, they must sum up
    to the ``data`` length. If they are floating numbers, they must sum up to 1.

    Parameters
    ----------
    data : sequence
        A sequence representing either SMILES strings or RDKit ``Mol`` objects.

    additional_data: sequence
        Additional sequences to be split alongside the main data, e.g. labels.

    train_size : float, default=None
        The fraction of data to be used for the train subset. If None, it is set
        to 1 - test_size - valid_size. If valid_size is not provided, train_size
        is set to 1 - test_size. If train_size, test_size and valid_size aren't
        set, train_size is set to 0.8.

    valid_size : float, default=None
        The fraction of data to be used for the test subset. If None, it is set
        to 1 - train_size - valid_size. If train_size, test_size and valid_size
        aren't set, train_size is set to 0.1.

    test_size : float, default=None
        The fraction of data to be used for the validation subset. If None, it is
        set to 1 - train_size - valid_size. If valid_size is not provided, test_size
        is set to 1 - train_size. If train_size, test_size and valid_size aren't set,
        test_size is set to 0.1.

    threshold : float, default=0.65
        Tanimoto distance threshold, defining the minimal distance between cluster
        centroids. Default value is based on ECFP4 activity threshold as determined
        by Roger Sayle [4]_.

    approximate : bool, default=False
        Whether to use approximate similarity calculation to speed up computation on
        large datasets. It uses NNDescent algorithm [5]_ [6]_ and requires `PyNNDescent`
        library to be installed. However, it is much slower on small datasets, and
        exact version is always used for data with less than 5000 molecules.

    return_indices : bool, default=False
        Whether the method should return the input object subsets, i.e. SMILES strings
        or RDKit ``Mol`` objects, or only the indices of the subsets instead of the data.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    Returns
    ----------
    subsets : tuple[list, list, ...]
        Tuple with train-valid-test subsets of provided arrays. First three are lists of
        SMILES strings or RDKit ``Mol`` objects, depending on the input type. If
        `return_indices` is True, lists of indices are returned instead of actual data.

    References
    ----------
    .. [1] `Darko Butina
        "Unsupervised Data Base Clustering Based on Daylight's Fingerprint and Tanimoto
        Similarity: A Fast and Automated Way To Cluster Small and Large Data Sets"
        . Chem. Inf. Comput. Sci. 1999, 39, 4, 747–750
        <https://pubs.acs.org/doi/abs/10.1021/ci9803381>`_

    .. [2] `Robin Taylor
        "Simulation Analysis of Experimental Design Strategies for Screening Random
        Compounds as Potential New Drugs and Agrochemicals"
        J. Chem. Inf. Comput. Sci. 1995, 35, 1, 59–67
        <https://pubs.acs.org/doi/10.1021/ci00023a009>`_

    .. [3] `Noel O'Boyle "Taylor-Butina Clustering"
        <https://noel.redbrick.dcu.ie/R_clustering.html>`_

    .. [4] `Roger Sayle
        "2D similarity, diversity and clustering in RDKit"
        RDKit UGM 2019
        <https://www.nextmovesoftware.com/talks/Sayle_2DSimilarityDiversityAndClusteringInRdkit_RDKITUGM_201909.pdf>`_

    .. [5] `W. Dong et al.
        "Efficient k-nearest neighbor graph construction for generic similarity measures"
        Proceedings of the 20th International World Wide Web Conference (WWW '11).
        Association for Computing Machinery, New York, NY, USA, 577–586
        <https://doi.org/10.1145/1963405.1963487>`_

    .. [6] `Leland McInnes
        "PyNNDescent for fast Approximate Nearest Neighbors"
        <https://pynndescent.readthedocs.io/en/latest/>`_
    """
    train_size, valid_size, test_size = validate_train_valid_test_split_sizes(
        train_size, valid_size, test_size, len(data)
    )

    clusters = _create_clusters(data, threshold, approximate, n_jobs)
    clusters.sort(key=len)

    train_idxs: list[int] = []
    valid_idxs: list[int] = []
    test_idxs: list[int] = []

    for cluster in clusters:
        if len(test_idxs) < test_size:
            test_idxs.extend(cluster)
        elif len(valid_idxs) < valid_size:
            valid_idxs.extend(cluster)
        else:
            train_idxs.extend(cluster)

    ensure_nonempty_subset(train_idxs, "train")
    ensure_nonempty_subset(valid_idxs, "validation")
    ensure_nonempty_subset(test_idxs, "test")

    if return_indices:
        train_subset = train_idxs
        valid_subset = valid_idxs
        test_subset = test_idxs
    else:
        train_subset = get_data_from_indices(data, train_idxs)
        valid_subset = get_data_from_indices(data, valid_idxs)
        test_subset = get_data_from_indices(data, test_idxs)

    if additional_data:
        additional_data_split: list[Sequence[Any]] = split_additional_data(
            list(additional_data), train_idxs, valid_idxs, test_idxs
        )
        return train_subset, valid_subset, test_subset, *additional_data_split
    else:
        return train_subset, valid_subset, test_subset


def _create_clusters(
    data: Sequence[Union[str, Mol]],
    threshold: float = 0.65,
    approximate: bool = False,
    n_jobs: Optional[int] = None,
) -> list[list[int]]:
    """
    Generate Taylor-Butina clusters for a list of SMILES strings or RDKit ``Mol`` objects.
    This function groups molecules by using clustering, where cluster centers must have
    Tanimoto (Jaccard) distance greater or equal to given threshold. Binary ECFP4 (Morgan)
    fingerprints with 2048 bits are used as features.
    """
    mols = ensure_mols(data)

    fps_rdkit = GetMorganGenerator().GetFingerprints(mols)
    centroid_idxs = LeaderPicker().LazyBitVectorPick(
        fps_rdkit, poolSize=len(mols), threshold=threshold
    )
    centroid_idxs = list(centroid_idxs)
    non_centroid_idxs = sorted(set(range(len(mols))) - set(centroid_idxs))

    # initially, each cluster is only its centroid
    clustering = {centroid_idx: [centroid_idx] for centroid_idx in centroid_idxs}
    clustering = defaultdict(list, clustering)

    # we don't use n_jobs here, since ECFP is too fast to benefit from that
    fps = ECFPFingerprint().transform(mols).astype(bool)
    fps_centroids = fps[centroid_idxs]
    fps_non_centroids = fps[non_centroid_idxs]

    # check nearest neighbors for the rest of the data and assign it to clusters
    # note that this is much faster, since Taylor-Butina results in a large number of
    # clusters, so we avoid a lot of nearest neighbor computations this way
    if not len(fps_non_centroids):
        # all points are their own centroids
        nearest_cluster_idxs = np.array([])
    elif not approximate or len(data) < 5000:
        nn = NearestNeighbors(n_neighbors=1, metric="jaccard", n_jobs=n_jobs)
        nn.fit(fps_centroids)
        nearest_cluster_idxs = nn.kneighbors(fps_non_centroids, return_distance=False)
    else:
        try:
            from pynndescent import NNDescent
        except ImportError:
            raise ImportError(
                "PyNNDescent not detected, which is needed for approximate Butina split. "
                "You can install it with: pip install pynndescent"
            )

        index = NNDescent(
            fps_centroids,
            metric="jaccard",
            random_state=0,
            parallel_batch_queries=True,
            n_jobs=n_jobs,
        )
        nearest_cluster_idxs, _ = index.query(fps_non_centroids, k=1)

    # assign rest of points to nearest neighbor clusters
    nearest_cluster_idxs = nearest_cluster_idxs.ravel()
    for mol_idx, cluster_idx in zip(non_centroid_idxs, nearest_cluster_idxs):
        clustering[cluster_idx].append(mol_idx)

    clusters = list(clustering.values())
    return clusters



from typing import Union

from rdkit.Chem import Crippen, Mol, rdMolDescriptors, rdmolops

from skfp.bases.base_filter import BaseFilter


class REOSFilter(BaseFilter):
    # flake8: noqa E501
    """
    REOS filter.

    REOS (Rapid Elimination Of Swill) is designed to filter out molecules with
    undesirable properties for drug discovery [1]_.

    Molecule must fulfill conditions:

    - molecular weight in range ``[200, 500]``
    - logP in range ``[-5, 5]``
    - HBA in range ``[0, 5]``
    - HBD in range ``[0, 10]``
    - charge in range ``[-2, 2]``
    - number of rotatable bonds in range ``[0, 8]``
    - number of heavy atoms in range ``[15, 50]``

    Parameters
    ----------
    allow_one_violation : bool, default=False
        Whether to allow violating one of the rules for a molecule. This makes the
        filter less restrictive.

    return_indicators : bool, default=False
        Whether to return a binary vector with indicators which molecules pass the
        filter, instead of list of molecules.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform_x_y` and
        :meth:`transform` are parallelized over the input molecules. ``None`` means 1
        unless in a :obj:`joblib.parallel_backend` context. ``-1`` means using all
        processors. See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when filtering molecules.

    References
    -----------
    .. [1] `Walters, W. P., Namchuk, M.
        "Designing screens: how to make your hits a hit"
        Nat Rev Drug Discov. 2003 Apr;2(4):259-66
        <https://pubmed.ncbi.nlm.nih.gov/12669025/>`_

    Examples
    ----------
    >>> from skfp.filters import REOSFilter
    >>> smiles = ["CC(C)CC1=CC=C(C=C1)C(C)C(=O)O",  "CC(=O)c1c(C(C)=O)c(C)n(CCCCn2c(C)c(C(C)=O)c(C(C)=O)c2C)c1C"]
    >>> filt = REOSFilter()
    >>> filt
    REOSFilter()
    >>> filtered_mols = filt.transform(smiles)
    >>> filtered_mols
    ['CC(C)CC1=CC=C(C=C1)C(C)C(=O)O']
    """

    def __init__(
        self,
        allow_one_violation: bool = False,
        return_indicators: bool = False,
        n_jobs: Union[int, None] = None,
        batch_size: Union[int, None] = None,
        verbose: int = 0,
    ):
        super().__init__(
            allow_one_violation, return_indicators, n_jobs, batch_size, verbose
        )

    def _apply_mol_filter(self, mol: Mol) -> bool:
        rules = [
            200 <= rdMolDescriptors.CalcExactMolWt(mol) <= 500,
            -5 <= Crippen.MolLogP(mol) <= 5,
            0 <= rdMolDescriptors.CalcNumHBA(mol) <= 10,
            0 <= rdMolDescriptors.CalcNumHBD(mol) <= 5,
            -2 <= rdmolops.GetFormalCharge(mol) <= 2,
            0 <= rdMolDescriptors.CalcNumRotatableBonds(mol) <= 8,
            15 <= rdMolDescriptors.CalcNumHeavyAtoms(mol) <= 50,
        ]

        passed_rules = sum(rules)

        if self.allow_one_violation:
            return passed_rules >= len(rules) - 1
        else:
            return passed_rules == len(rules)



from collections import defaultdict
from collections.abc import Sequence
from copy import deepcopy
from numbers import Integral
from typing import Any, Optional, Union

from rdkit import Chem
from rdkit.Chem import Mol
from rdkit.Chem.Scaffolds import MurckoScaffold
from sklearn.utils._param_validation import Interval, RealNotInt, validate_params

from skfp.model_selection.splitters.utils import (
    ensure_nonempty_subset,
    split_additional_data,
    validate_train_test_split_sizes,
    validate_train_valid_test_split_sizes,
)
from skfp.utils.functions import get_data_from_indices
from skfp.utils.validators import ensure_mols


@validate_params(
    {
        "data": ["array-like"],
        "additional_data": ["tuple"],
        "train_size": [
            Interval(RealNotInt, 0, 1, closed="neither"),
            Interval(Integral, 1, None, closed="left"),
            None,
        ],
        "test_size": [
            Interval(RealNotInt, 0, 1, closed="neither"),
            Interval(Integral, 1, None, closed="left"),
            None,
        ],
        "use_csk": ["boolean"],
        "return_indices": ["boolean"],
    },
    prefer_skip_nested_validation=True,
)
def scaffold_train_test_split(
    data: Sequence[Union[str, Mol]],
    *additional_data: Sequence,
    train_size: Optional[float] = None,
    test_size: Optional[float] = None,
    use_csk: bool = False,
    return_indices: bool = False,
) -> Union[
    tuple[
        Sequence[Union[str, Mol]], Sequence[Union[str, Mol]], Sequence[Sequence[Any]]
    ],
    tuple[Sequence, ...],
    tuple[Sequence[int], Sequence[int]],
]:
    """
    Split using groups of Bemis-Murcko scaffolds.

    This split uses deterministically partitioned groups of Bemis-Murcko molecular
    scaffolds [1]_ for splitting, as introduced in the MoleculeNet [2]_ paper. It aims
    to verify the model generalization to new and rare scaffolds, as an approximation
    to the time split.

    By default, core structure scaffolds are used (following RDKit), which include atom
    types. Original Bemis-Murcko approach uses the cyclic skeleton of a molecule, replacing
    all atoms by carbons. It is also known as CSK (Cyclic SKeleton) [3]_, and can be
    used with `use_csk` parameter.

    This approach is known to have certain limitations. In particular, molecules with
    no rings will not get a scaffold, resulting in them being grouped together regardless
    of their structure.

    The split is fully deterministic, with the smallest scaffold sets assigned to the test
    subset and the rest to the training subset.

    If ``train_size`` and ``test_size`` are integers, they must sum up to the ``data``
    length. If they are floating numbers, they must sum up to 1.

    Parameters
    ----------
    data : sequence
        A sequence representing either SMILES strings or RDKit ``Mol`` objects.

    additional_data: list[sequence]
        Additional sequences to be split alongside the main data (e.g., labels or feature vectors).

    train_size : float, default=None
        The fraction of data to be used for the train subset. If None, it is set to 1 - test_size.
        If test_size is also None, it will be set to 0.8.

    test_size : float, default=None
        The fraction of data to be used for the test subset. If None, it is set to 1 - train_size.
        If train_size is also None, it will be set to 0.2.

    use_csk: bool, default=False
        Whether to use molecule's skeleton or the core structure scaffold (including atom types).

    return_indices : bool, default=False
        Whether the method should return the input object subsets, i.e. SMILES strings
        or RDKit ``Mol`` objects, or only the indices of the subsets instead of the data.

    Returns
    ----------
    subsets : tuple[list, list, ...]
        Tuple with train-test subsets of provided arrays. First two are lists of SMILES
        strings or RDKit ``Mol`` objects, depending on the input type. If `return_indices`
        is True, lists of indices are returned instead of actual data.

    References
    ----------
    .. [1] `Bemis, G. W., & Murcko, M. A.
        "The properties of known drugs. 1. Molecular frameworks."
        Journal of Medicinal Chemistry, 39(15), 2887-2893.
        <https://www.researchgate.net/publication/14493474_The_Properties_of_Known_Drugs_1_Molecular_Frameworks>`_

    .. [2] `Z. Wu, B. Ramsundar, E. N. Feinberg, J. Gomes, C. Geniesse, A. S. Pappu, K. Leswing, V. Pande
        "MoleculeNet: A Benchmark for Molecular Machine Learning."
        Chemical Science, 9(2), 513-530.
        <https://www.researchgate.net/publication/314182452_MoleculeNet_A_Benchmark_for_Molecular_Machine_Learning>`_

    .. [3] ` Bemis-Murcko scaffolds and their variants
        <https://github.com/rdkit/rdkit/discussions/6844>`_
    """
    train_size, test_size = validate_train_test_split_sizes(
        train_size, test_size, len(data)
    )

    scaffold_sets = _create_scaffold_sets(data, use_csk)
    scaffold_sets.sort(key=len)

    train_idxs: list[int] = []
    test_idxs: list[int] = []

    for scaffold_set in scaffold_sets:
        if len(test_idxs) < test_size:
            test_idxs.extend(scaffold_set)
        else:
            train_idxs.extend(scaffold_set)

    ensure_nonempty_subset(train_idxs, "train")
    ensure_nonempty_subset(test_idxs, "test")

    if return_indices:
        train_subset = train_idxs
        test_subset = test_idxs
    else:
        train_subset = get_data_from_indices(data, train_idxs)
        test_subset = get_data_from_indices(data, test_idxs)

    if additional_data:
        additional_data_split: list[Sequence[Any]] = split_additional_data(
            list(additional_data), train_idxs, test_idxs
        )
        return train_subset, test_subset, *additional_data_split
    else:
        return train_subset, test_subset


@validate_params(
    {
        "data": ["array-like"],
        "additional_data": ["tuple"],
        "train_size": [
            Interval(RealNotInt, 0, 1, closed="neither"),
            Interval(Integral, 1, None, closed="left"),
            None,
        ],
        "valid_size": [
            Interval(RealNotInt, 0, 1, closed="neither"),
            Interval(Integral, 1, None, closed="left"),
            None,
        ],
        "test_size": [
            Interval(RealNotInt, 0, 1, closed="neither"),
            Interval(Integral, 1, None, closed="left"),
            None,
        ],
        "use_csk": ["boolean"],
        "return_indices": ["boolean"],
    },
    prefer_skip_nested_validation=True,
)
def scaffold_train_valid_test_split(
    data: Sequence[Union[str, Mol]],
    *additional_data: Sequence,
    train_size: Optional[float] = None,
    valid_size: Optional[float] = None,
    test_size: Optional[float] = None,
    use_csk: bool = False,
    return_indices: bool = False,
) -> Union[
    tuple[
        Sequence[Union[str, Mol]],
        Sequence[Union[str, Mol]],
        Sequence[Union[str, Mol]],
        Sequence[Sequence[Any]],
    ],
    tuple[Sequence, ...],
    tuple[Sequence[int], Sequence[int], Sequence[int]],
]:
    """
    Split using groups of Bemis-Murcko scaffolds.

    This split uses deterministically partitioned groups of Bemis-Murcko molecular
    scaffolds [1]_ for splitting, as introduced in the MoleculeNet [2]_ paper. It aims
    to verify the model generalization to new and rare scaffolds, as an approximation
    to the time split.

    By default, core structure scaffolds are used (following RDKit), which include atom
    types. Original Bemis-Murcko approach uses the cyclic skeleton of a molecule, replacing
    all atoms by carbons. It is also known as CSK (Cyclic SKeleton) [3]_, and can be
    used with `use_csk` parameter.

    This approach is known to have certain limitations. In particular, molecules with
    no rings will not get a scaffold, resulting in them being grouped together regardless
    of their structure.

    The split is fully deterministic, with the smallest scaffold sets assigned to the test
    subset, larger to the validation subset, and the rest to the training subset.

    If ``train_size``, ``valid_size`` and ``test_size`` are integers, they must sum up
    to the ``data`` length. If they are floating numbers, they must sum up to 1.

    Parameters
    ----------
    data : sequence
        A sequence representing either SMILES strings or RDKit ``Mol`` objects.

    additional_data: sequence
        Additional sequences to be split alongside the main data, e.g. labels.

    train_size : float, default=None
        The fraction of data to be used for the train subset. If None, it is set
        to 1 - test_size - valid_size. If valid_size is not provided, train_size
        is set to 1 - test_size. If train_size, test_size and valid_size aren't
        set, train_size is set to 0.8.

    valid_size : float, default=None
        The fraction of data to be used for the test subset. If None, it is set
        to 1 - train_size - valid_size. If train_size, test_size and valid_size
        aren't set, train_size is set to 0.1.

    test_size : float, default=None
        The fraction of data to be used for the validation subset. If None, it is
        set to 1 - train_size - valid_size. If valid_size is not provided, test_size
        is set to 1 - train_size. If train_size, test_size and valid_size aren't set,
        test_size is set to 0.1.

    use_csk: bool, default=False
        Whether to use the molecule cyclic skeleton (CSK), instead of the core
        structure scaffold.

    return_indices : bool, default=False
        Whether the method should return the input object subsets, i.e. SMILES strings
        or RDKit ``Mol`` objects, or only the indices of the subsets instead of the data.

    Returns
    ----------
    subsets : tuple[list, list, ...]
        Tuple with train-valid-test subsets of provided arrays. First three are lists of
        SMILES strings or RDKit ``Mol`` objects, depending on the input type. If
        `return_indices` is True, lists of indices are returned instead of actual data.

    References
    ----------
    .. [1] `Bemis, G. W., & Murcko, M. A.
        "The properties of known drugs. 1. Molecular frameworks."
        Journal of Medicinal Chemistry, 39(15), 2887-2893.
        <https://www.researchgate.net/publication/14493474_The_Properties_of_Known_Drugs_1_Molecular_Frameworks>`_

    .. [2] `Z. Wu, B. Ramsundar, E. N. Feinberg, J. Gomes, C. Geniesse, A. S. Pappu, K. Leswing, V. Pande
        "MoleculeNet: A Benchmark for Molecular Machine Learning."
        Chemical Science, 9(2), 513-530.
        <https://www.researchgate.net/publication/314182452_MoleculeNet_A_Benchmark_for_Molecular_Machine_Learning>`_

    .. [3] ` Bemis-Murcko scaffolds and their variants
        <https://github.com/rdkit/rdkit/discussions/6844>`_
    """
    train_size, valid_size, test_size = validate_train_valid_test_split_sizes(
        train_size, valid_size, test_size, len(data)
    )

    scaffold_sets = _create_scaffold_sets(data, use_csk)
    scaffold_sets.sort(key=len)

    train_idxs: list[int] = []
    valid_idxs: list[int] = []
    test_idxs: list[int] = []

    for scaffold_set in scaffold_sets:
        if len(test_idxs) < test_size:
            test_idxs.extend(scaffold_set)
        elif len(valid_idxs) < valid_size:
            valid_idxs.extend(scaffold_set)
        else:
            train_idxs.extend(scaffold_set)

    ensure_nonempty_subset(train_idxs, "train")
    ensure_nonempty_subset(valid_idxs, "validation")
    ensure_nonempty_subset(test_idxs, "test")

    if return_indices:
        train_subset = train_idxs
        valid_subset = valid_idxs
        test_subset = test_idxs
    else:
        train_subset = get_data_from_indices(data, train_idxs)
        valid_subset = get_data_from_indices(data, valid_idxs)
        test_subset = get_data_from_indices(data, test_idxs)

    if additional_data:
        additional_data_split: list[Sequence[Any]] = split_additional_data(
            list(additional_data), train_idxs, valid_idxs, test_idxs
        )
        return train_subset, valid_subset, test_subset, *additional_data_split
    else:
        return train_subset, valid_subset, test_subset


def _create_scaffold_sets(
    data: Sequence[Union[str, Mol]], use_csk: bool = False
) -> list[list[int]]:
    """
    Generate Bemis-Murcko scaffolds for a list of SMILES strings or RDKit ``Mol`` objects.
    This function groups molecules by their Bemis-Murcko scaffold into sets of molecules
    with the same scaffold.

    They can be generated as either the core structure scaffold (with atom types) or the
    skeleton scaffold (without atom types).
    """
    scaffold_sets = defaultdict(list)
    mols = ensure_mols(data)

    for idx, mol in enumerate(mols):
        mol = deepcopy(mol)
        Chem.RemoveStereochemistry(mol)

        if use_csk:
            scaffold = MurckoScaffold.GetScaffoldForMol(mol)
            scaffold = MurckoScaffold.MakeScaffoldGeneric(scaffold)
            scaffold = MurckoScaffold.GetScaffoldForMol(scaffold)
        else:
            scaffold = MurckoScaffold.MurckoScaffoldSmiles(mol=mol)

        scaffold_sets[scaffold].append(idx)

    scaffold_sets = list(scaffold_sets.values())
    return scaffold_sets



import logging
from collections.abc import Sequence
from numbers import Integral, Real
from typing import Optional, Union

import mmh3
import numpy as np
import scipy.sparse
from rdkit import RDLogger
from rdkit.Chem import Mol, MolToSmiles
from scipy.sparse import csr_array
from sklearn.utils._param_validation import Interval, InvalidParameterError

from skfp.bases import BaseFingerprintTransformer
from skfp.utils import require_mols_with_conf_ids

"""
Note: this file cannot have the "e3fp.py" name due to conflict with E3FP library.
"""

# e3fp library has a problem with mmh3 version 5.x, so we monkey-patch it
import e3fp.fingerprint.fprinter

fixed_hash = lambda array, seed=0: mmh3.hash(array.tobytes(), seed)
e3fp.fingerprint.fprinter.hash_int64_array = fixed_hash


class E3FPFingerprint(BaseFingerprintTransformer):
    """
    E3FP (Extended 3-Dimensional FingerPrint) fingerprint.

    The implementation uses ``e3fp`` library. This is a hashed fingerprint [1]_, where
    fragments are computed based on "shells", i.e. spherical areas around each
    atom in the 3D conformation of a molecule. The initial vector is quite large,
    and is then folded to the ``fp_size`` length.

    Shells are created around each atom with increasing radius, multiplied each
    time by ``radius_multiplier``, until ``level`` iterations are reached or when there
    is no change. Shell of each radius is hashed.

    Each shells get an identifier based on atom types in their radius, which is then
    hashed. Atom types (invariants) by default are based on Daylight invariants:

    - number of heavy neighbors
    - valence (excluding hydrogen neighbors)
    - atomic number
    - atomic mass
    - formal charge
    - number of bound hydrogens
    - whether it is a part of a ring

    This is a 3D fingerprint, and requries molecules with ``conf_id`` integer property
    set. They can be generated with :class:`~skfp.preprocessing.ConformerGenerator`.

    Parameters
    ----------
    fp_size : int, default=1024
        Size of output vectors, i.e. number of bits for each fingerprint. Must be
        positive.

    n_bits_before_folding : int, default=4096
        Size of fingerprint vector after initial hashing. It is then folded to
        ``fp_size``. Must be positive, and larger or equal to ``fp_size``.

    level : int, default=None
        Maximal number of iterations with increasing shell radius. None means that
        it stops only when the is no change from the last iteration.

    radius_multiplier : float, default=1.718
        How much to multiply the radius by to get the next, larger shell. Must be
        greater than 1.

    rdkit_invariants : bool, default=False
        Whether to use RDKit ECFP invariants instead of Daylight ones.

    count : bool, default=False
        Whether to return binary (bit) features, or their counts.

    sparse : bool, default=False
        Whether to return dense NumPy array, or sparse SciPy CSR array.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    random_state : int, RandomState instance or None, default=0
        Controls the randomness of conformer generation.

    Attributes
    ----------
    n_features_out : int
        Number of output features, size of fingerprints. Equal to ``fp_size``.

    requires_conformers : bool = True
        Value is always True, as this fingerprint is 3D based. It always requires
        molecules with conformers as inputs, with ``conf_id`` integer property set.

    See Also
    --------
    :class:`ECFPFingerprint` : Related 2D fingerprint. E3FP was designed to extend it
        to 3D features.

    References
    ----------
    .. [1] `Axen, Seth D., et al.
        "A simple representation of three-dimensional molecular structure"
        J. Med. Chem. 2017, 60, 17, 7393–7409
        <https://pubs.acs.org/doi/abs/10.1021/acs.jmedchem.7b00696>`_

    Examples
    --------
    >>> from skfp.fingerprints import E3FPFingerprint
    >>> from skfp.preprocessing import MolFromSmilesTransformer, ConformerGenerator
    >>> smiles = ["O", "CC", "[C-]#N", "CC=O"]
    >>> fp = E3FPFingerprint()
    >>> fp
    E3FPFingerprint()

    >>> mol_from_smiles = MolFromSmilesTransformer()
    >>> mols = mol_from_smiles.transform(smiles)
    >>> conf_gen = ConformerGenerator()
    >>> mols = conf_gen.transform(mols)
    >>> fp.transform(mols)
    array([[0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0]], dtype=uint8)
    """

    _parameter_constraints: dict = {
        **BaseFingerprintTransformer._parameter_constraints,
        "fp_size": [Interval(Integral, 1, None, closed="left")],
        "n_bits_before_folding": [Interval(Integral, 1, None, closed="left")],
        "level": [None, Interval(Integral, 1, None, closed="left")],
        "radius_multiplier": [Interval(Real, 1.0, None, closed="neither")],
        "rdkit_invariants": ["boolean"],
    }

    def __init__(
        self,
        fp_size: int = 1024,
        n_bits_before_folding: int = 4096,
        level: Optional[int] = None,
        radius_multiplier: float = 1.718,
        rdkit_invariants: bool = False,
        count: bool = False,
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
        random_state: Optional[int] = 0,
    ):
        super().__init__(
            n_features_out=fp_size,
            requires_conformers=True,
            count=count,
            sparse=sparse,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
            random_state=random_state,
        )
        self.fp_size = fp_size
        self.n_bits_before_folding = n_bits_before_folding
        self.level = level
        self.radius_multiplier = radius_multiplier
        self.rdkit_invariants = rdkit_invariants

    def _validate_params(self) -> None:
        super()._validate_params()
        if self.n_bits_before_folding < self.fp_size:
            raise InvalidParameterError(
                f"n_bits_before_folding must be greater of equal to fp_size, got:"
                f"n_bits_before_folding={self.n_bits_before_folding}, "
                f"fp_size={self.fp_size}"
            )

    def transform(
        self, X: Sequence[Union[str, Mol]], copy: bool = False
    ) -> Union[np.ndarray, csr_array]:
        """
        Compute E3FP fingerprints.

        Parameters
        ----------
        X : {sequence, array-like} of shape (n_samples,)
            Sequence containing RDKit Mol objects, with conformers generated and
            ``conf_id`` integer property set.

        copy : bool, default=False
            Copy the input X or not.

        Returns
        -------
        X : {ndarray, sparse matrix} of shape (n_samples, self.fp_size)
            Array with fingerprints.
        """
        return super().transform(X, copy)

    def _calculate_fingerprint(self, X: Sequence[Mol]) -> Union[np.ndarray, csr_array]:
        X = require_mols_with_conf_ids(X)
        X = [self._calculate_single_mol_fingerprint(mol) for mol in X]
        return scipy.sparse.vstack(X) if self.sparse else np.array(X)

    def _calculate_single_mol_fingerprint(
        self, mol: Mol
    ) -> Union[np.ndarray, csr_array]:
        from e3fp.pipeline import fprints_from_mol

        # e3fp requires "_Name" property to be set
        mol.SetProp("_Name", MolToSmiles(mol))

        # suppress flood of logs
        try:
            if not self.verbose:
                logging.disable(logging.INFO)
                RDLogger.DisableLog("rdApp.*")

            fps = fprints_from_mol(
                mol,
                fprint_params={
                    "bits": self.n_bits_before_folding,
                    "level": self.level,
                    "radius_multiplier": self.radius_multiplier,
                    "rdkit_invariants": self.rdkit_invariants,
                    "counts": self.count,
                },
            )
        finally:
            RDLogger.EnableLog("rdApp.*")
            logging.disable(logging.NOTSET)

        mol.ClearProp("_Name")

        fp = fps[0]
        fp = fp.fold(self.fp_size)
        dtype = np.uint32 if self.count else np.uint8
        return fp.to_vector(sparse=self.sparse, dtype=dtype)
